//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
program lossyWAV;
//==============================================================================
// Conditional Defines
//==============================================================================
{$R *.RES}
//==============================================================================
uses
  fftw_interface in 'units\fftw_interface.pas',
  {$IFDEF USEASM}
  nFFT_asm in 'units\nFFT_asm.pas',
  {$ELSE}
  nFFT_pas in 'units\nFFT_pas.pas',
  {$ENDIF}
  nFillFFT in 'units\nFillFFT.pas',
  nInitialise in 'units\nInitialise.pas',
  nMaths in 'units\nMaths.pas',
  nOutput in 'units\nOutput.pas',
  nParameter in 'units\nParameter.pas',
  nProcess in 'units\nProcess.pas',
  nRemoveBits in 'units\nRemoveBits.pas',
  nShiftBlocks in 'units\nShiftBlocks.pas',
  nSpreading in 'units\nSpreading.pas',
  nSystem in 'units\nSystem.pas',
  nWav in 'units\nWav.pas',
  Sysutils,
  Math;

//==============================================================================
begin
//==============================================================================
  Check_Switches();

  nOutput_Init;

  if merging then
    MergeFiles()
  else
  begin
    if not openWavIO then
      lossyWAVerror('Error initialising wavIO unit.',$11);

    if Global.codec_block_size=0 then
      lossyWAVerror('Error initialising wavIO unit.',$11);

    Initial_Setup();

    Skewing_Function_Init();

    Fill_FFT_Lookup_Init(); //dependent on codec_block_size.

    if not readNextNextCodecBlock() then
      lossyWAVerror('Error reading from input file.',$21);

    Global.blocks_processed:=0;

    //==========================================================================
    // Main processing loop.
    //==========================================================================
    while (next_codec_block_size>0) do
    begin
      shift_codec_blocks();

      readNextNextCodecBlock();

      process_this_codec_block();

      if not writeNextBtrdCodecBlock() then
        lossyWAVerror('Error writing to output file.',$21);

      if correct and not writeNextCorrCodecBlock() then
        lossyWAVerror('Error writing to correction file.',$22);
    //==========================================================================
    end;
    //==========================================================================

    if not CloseWavIO() then
      lossyWAVerror('Error closing wavIO unit.',$11);

    Write_Cleanup();
  end;
//==============================================================================
end.
//==============================================================================

