{--------------------------------------------------------------}
{ Delphi interface to the FFTW library -- FFTW Version 3.0.1.  }
{ Note that this interface is incomplete. Additional function  }
{ interface entries may be added in an anologous manner, see   }
{ fftw  for more details.                  }
{                                                              }
{ Last modified 22/DEC/03                                      }
{ Written by Mark G. Beckett (g.beckett@epcc.ed.ac.uk          }
{--------------------------------------------------------------}
{ Modified 07/JUN/2009                                         }
{ by Nick Currie (lossyWAV at yahoo dot com)                   }
{ Now dynamically links rather than statically. Required to    }
{ stop program crashing if libfftw3-3.DLL is not available.    }
{--------------------------------------------------------------}

//==============================================================================
unit fftw_interface;
//==============================================================================
interface
 {EPCC (MGB) - Note that the FFTW memory allocation routines can not be used
  for data arrays, because the DLL is allocated a separate heap that is not
  accessible from the Delphi runtime environment.}
//==============================================================================
uses
  nSystem;
//==============================================================================
var
  Unit_Data_Packer1       : Array[0..15] of Integer;
//==============================================================================
const
 {EPCC (MGB) - FFTW documented constants, taken from "api/fftw-3.h". Comments
  to the right of the definitions are transcribed from the original header
  file.}

  FFTW_FORWARD            = -1;
  FFTW_BACKWARD           = 1;

  FFTW_MEASURE            = 0;
  FFTW_DESTROY_INPUT      = 1 shl 0; {01h; 1U << 0}
  FFTW_UNALIGNED          = 1 shl 1; {02h; 1U << 1}
  FFTW_CONSERVE_MEMORY    = 1 shl 2; {04h; 1U << 2}
  FFTW_EXHAUSTIVE         = 1 shl 3; {08h; 1U << 3} {NO_EXHAUSTIVE is default }
  FFTW_PRESERVE_INPUT     = 1 shl 4; {10h; 1U << 4} {cancels FFTW_DESTROY_INPUT}
  FFTW_PATIENT            = 1 shl 5; {20h; 1U << 5} {IMPATIENT is default }
  FFTW_ESTIMATE           = 1 shl 6; {40h; 1U << 6}

 {FFTW undocumented constants have not been defined in this implementation.
  They are not required for typical usage of the library.}

//==============================================================================
var
  fftw_plan_dft_r2c_1d  : function (n : Integer; inData : PDouble; outData : PDouble; flags : Longword) : Pointer; cdecl;
  fftw_execute          : Procedure(plan : Pointer); cdecl;
  fftw_destroy_plan     : Procedure(plan : Pointer); cdecl;

  fftw_plans            : array[0..MAX_FFT_BIT_LENGTH] of pointer;
//==============================================================================

function  FFTW_Initialised : Longbool;
procedure FFTW_Cleanup;

//==============================================================================
implementation
//==============================================================================
uses
  Windows;
//==============================================================================
var
  FFTWDLL_Initialised   : LongBool;
  FFTWDLL_Handle        : THandle;
  Unit_Data_Packer2     : Array[0..11] of Integer;
//==============================================================================

//==============================================================================
Procedure FFTW_initialise;
//==============================================================================
begin
  FFTWDLL_Initialised:=False;
  FFTWDLL_Handle:=LoadLibrary('libfftw3-3.dll');
  if FFTWDLL_Handle<>0 then
  begin
    FFTWDLL_Initialised :=True;
    fftw_plan_dft_r2c_1d:=GetProcAddress(FFTWDLL_Handle,'fftw_plan_dft_r2c_1d');
    fftw_execute        :=GetProcAddress(FFTWDLL_Handle,'fftw_execute');
    fftw_destroy_plan   :=GetProcAddress(FFTWDLL_Handle,'fftw_destroy_plan');
  end;
  If (FFTWDLL_Handle=0) or (@fftw_plan_dft_r2c_1d=Nil) or (@fftw_execute=Nil) or (@fftw_destroy_plan=Nil) then
    FFTWDLL_Initialised:=False;
end;
//==============================================================================

//==============================================================================
Function FFTW_Initialised : Longbool;
//==============================================================================
begin
  FFTW_Initialised:=FFTWDLL_Initialised;
end;
//==============================================================================

//==============================================================================
procedure FFTW_Cleanup;
//==============================================================================
begin
  If FFTWDLL_Initialised then
  begin
    FreeLibrary(FFTWDLL_Handle);
    FFTWDLL_Handle:=0;
    FFTWDLL_Initialised:=False;
    fftw_plan_dft_r2c_1d  :=Nil;
    fftw_execute          :=Nil;
    fftw_destroy_plan     :=Nil;
  end;
end;
//==============================================================================
Initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('FFTW_Interface',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}

  FFTW_Initialise();
//==============================================================================
end.
//==============================================================================
