//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nFillFFT;
//==============================================================================
interface
//==============================================================================
uses
  nwav,
  nparameter,
  nSystem;
type
  DoubleArray = array[0..0] of double;
//==============================================================================
var
  Unit_Data_Packer1                     : Array[0..15] of Integer;

  //============================================================================
  // window function lookup table.
  //============================================================================
  window_function                       : array[0..MAX_FFT_BIT_LENGTH] of ^DoubleArray;
  window_function_rms                   : Double;
  Fill_FFT_Input_log2RMS                : Double;

  codec_block_fill_offset               : Integer;
  actual_analysis_block_start           : Integer;

//==============================================================================

Procedure Fill_FFT_Lookup_Init();
Function  Fill_FFT_Input() : Extended;

//==============================================================================
implementation
//==============================================================================
uses
  nmaths,
  math;
//==============================================================================
var
  //============================================================================
  // Fill_FFT_Input procedure data lookup table.
  //============================================================================
  fill_fft_lookup                       : array[0..MAX_BLOCK_SIZE*4-1] of packed record
                                            block  : integer;
                                            offset : integer;
                                          end;
  {$IFDEF USEASM}
  Unit_Data_Packer2 : Array[0..0] of integer;
  {$ELSE}
  Unit_Data_Packer2 : Array[0..2] of integer;
  {$ENDIF}
//==============================================================================

  {$IFNDEF USEASM}
  //----------------------------------------------------------------------------
  Procedure Fill_FFT_Lookup_Init;
  //----------------------------------------------------------------------------
  Var
    ff_i : Integer;
  //----------------------------------------------------------------------------
  Begin
    for ff_i:=0 to Global.codec_block_size*4-1 do
      with fill_fft_lookup[ff_i] do
      begin
        block:=(ff_i div Global.codec_block_size);
        offset:=(ff_i mod Global.codec_block_size);
      end;
    codec_block_fill_offset:=Global.codec_block_size*2;
  //----------------------------------------------------------------------------
  End;
  //----------------------------------------------------------------------------

  //----------------------------------------------------------------------------
  function Fill_FFT_Input : extended;
  //----------------------------------------------------------------------------
  var
    ff_i, ff_j, ff_n : Integer;
    ff_k, ff_m : Double;
  //----------------------------------------------------------------------------
  begin
    ff_k:=0;
    ff_n:=codec_block_fill_offset+actual_analysis_block_start;
    for ff_i:=0 to Global.fft_length-1 do
    begin
      ff_j:=ff_i+ff_n;
      ff_m:=rotating_blocks_ptr[fill_fft_lookup[ff_j].block]^[Global.channel,fill_fft_lookup[ff_j].offset];
      ff_k:=ff_k+ff_m*ff_m;
      {$R-}
      fft_array.Real[ff_i]:=ff_m*window_function[Global.fft_bit_length]^[ff_i];
      {$R+}
    end;
    Fill_FFT_Input_log2RMS:=fastl2root(ff_k*Global.fft_length_recip);
    Fill_FFT_Input:=Fill_FFT_Input_log2RMS;
  end;
  //----------------------------------------------------------------------------
  {$ELSE}
  //============================================================================
  Procedure Fill_FFT_Lookup_Init;
  //============================================================================
  Asm
                push    ebx
                push    edi

                mov     eax,Global.codec_block_size
                shl     eax,1
                mov     codec_block_fill_offset,eax
                shl     eax,1

                jmp     @@02

      @@01:     mov     ecx,Global.codec_block_bits
                mov     edx,dword ptr PowersOf.twom1[ecx*int64_size]

                mov     ebx,eax
                mov     edi,eax
                and     edi,edx
                shr     ebx,cl

                mov     dword ptr fill_fft_lookup[eax*(INTEGER_SIZE*2)],ebx
                mov     dword ptr fill_fft_lookup[eax*(INTEGER_SIZE*2)+INTEGER_SIZE],edi

      @@02:     sub     eax,1
                jge     @@01

                pop     edi
                pop     ebx
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  function Fill_FFT_Input : extended;
  //============================================================================
  asm
                push    ebx
                push    edi
                push    esi

                sub     edi,edi
                sub     edx,edx
                fldz

                mov     ecx,Global.fft_bit_length
                mov     ecx,dword ptr window_function[ecx*POINTER_SIZE]

                mov     eax,Global.channel
                mov     eax,dword ptr CHANNEL_OFFSET[eax*POINTER_SIZE]

                mov     ebx,codec_block_fill_offset
                add     ebx,actual_analysis_block_start
                shl     ebx,(integer_shift+1)

      @@00:       mov     edx,dword ptr fill_fft_lookup[ebx+edi*(2*INTEGER_SIZE)]
                  mov     edx,dword ptr rotating_blocks_ptr[edx*POINTER_SIZE]
                  add     edx,eax

                  mov     esi,dword ptr fill_fft_lookup[ebx+edi*(2*INTEGER_SIZE)+INTEGER_SIZE]

                  fild    dword ptr [edx+esi*INTEGER_SIZE]
                  fld     st

                  fmul    qword ptr [ecx+edi*DOUBLE_SIZE]
                  fstp    qword ptr FFT_Array[edi*DOUBLE_SIZE]

                  fmul    st,st
                  fadd

                  add     edi,1
                  cmp     edi,Global.fft_length
                  jl      @@00

                call    fastlog2root
                fisub   GLOBAL.fft_bit_length

                pop     esi
                pop     edi
                pop     ebx
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  //============================================================================
  Procedure Window_Function_Lookup_Init(this_alpha : Double);
  //============================================================================
  var
    ff_i, ff_j : Integer;
  //============================================================================
  begin
    // Generate window_function lookup tables
    for ff_j:=0 to MAX_FFT_BIT_LENGTH do
    begin
      getmem(window_function[ff_j],powersof.twoint[ff_j+3]);
      for ff_i:=0 to PowersOf.twom1[ff_j] do
      {$R-}
        window_function[ff_j]^[ff_i]:=hanning(ff_i,ff_j);
      {$R+}
    end;
    window_function_rms:=-0.70751875638961792*log10_2x20;     // == RMS of Hanning Window x 20 x log10(2).
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure window_function_lookup_final;
  //============================================================================
  var
    ff_i : integer;
  begin
    for ff_i:=0 to max_fft_bit_length do
      freemem(window_function[ff_i]);
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('nFillFFT',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}

  Window_Function_Lookup_Init(0);
//==============================================================================

//==============================================================================
finalization
//==============================================================================
  Window_Function_Lookup_Final;
//==============================================================================

//==============================================================================
end.
//==============================================================================
