//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nInitialise;
//==============================================================================
interface
//==============================================================================
var
  Unit_Data_Packer1 : Array[0..15] of integer;
//==============================================================================
Procedure Check_Switches;
procedure initial_setup;
//==============================================================================
implementation
//==============================================================================
uses
  fftw_interface,
  nFillFFT,
  nmaths,
  nparameter,
  nremovebits,
  nspreading,
  nsystem,
  nwav,
  math,
  sysutils;
//==============================================================================
const
//==============================================================================
  QUALITY_PRESET_MIN                    = -4;
  QUALITY_PRESET_MAX                    = 10;
  //============================================================================
  // constants directly related to quality_integer settings.
  //============================================================================                    //    -4,    -3,    -2,    -1,     0,     1,     2,     3,     4,     5,     6,     7,     8,     9,    10
  QUALITY_MINIMUM_BITS_TO_KEEP          : array[0                 ..QUALITY_PRESET_MAX] of Single   = (                              2.50,  2.75,  3.00,  3.25,  3.50,  3.75,  4.00,  4.25,  4.50,  4.75,  5.00);
//QUALITY_MINIMUM_BITS_TO_KEEP_ALT      : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (  2.50,  2.60,  2.70,  2.80,  2.90,  3.00,  3.10,  3.20,  3.30,  3.40,  3.50,  3.60,  3.70,  3.80,  3.90); //1.1.5b
  QUALITY_MINIMUM_BITS_TO_KEEP_ALT      : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (  2.50,  2.64,  2.78,  2.92,  3.06,  3.19,  3.33,  3.47,  3.61,  3.75,  4.00,  4.25,  4.50,  4.75,  5.00); //1.1.5c
                                                                                                    //    -4,    -3,    -2,    -1,     0,     1,     2,     3,     4,     5,     6,     7,     8,     9,    10
  QUALITY_NOISE_THRESHOLD_SHIFTS        : array[0                 ..QUALITY_PRESET_MAX] of Single   = (                             20.00, 16.00,  9.00,  6.00,  3.00,  0.00, -2.40, -4.80, -7.20, -9.60,-12.00);
//QUALITY_NOISE_THRESHOLD_SHIFTS_ALT    : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = ( 20.00, 17.75, 15.50, 13.25, 11.00,  8.75,  6.50,  4.25,  2.00, -0.25, -2.50, -4.75, -7.00, -9.25,-11.50); //b
  QUALITY_NOISE_THRESHOLD_SHIFTS_ALT    : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = ( 20.00, 17.78, 15.56, 13.33, 11.11,  8.89,  6.67,  4.44,  2.22,  0.00, -2.22, -4.44, -6.67, -8.89,-11.11); //c
                                                                                                    //    -4,    -3,    -2,    -1,     0,     1,     2,     3,     4,     5,     6,     7,     8,     9,    10
  QUALITY_SIGNAL_TO_NOISE_RATIOS        : array[0                 ..QUALITY_PRESET_MAX] of Single   = (                            -18.00,-22.00,-23.50,-23.50,-23.50,-25.00,-28.00,-31.00,-34.00,-37.00,-40.00);
//QUALITY_SIGNAL_TO_NOISE_RATIOS_ALT    : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (-18.00,-19.00,-20.00,-21.00,-22.00,-23.00,-24.00,-25.00,-26.00,-27.00,-28.00,-29.00,-30.00,-31.00,-32.00); //b
  QUALITY_SIGNAL_TO_NOISE_RATIOS_ALT    : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (-18.00,-19.00,-20.00,-21.00,-22.00,-23.00,-24.00,-25.00,-26.00,-27.00,-28.00,-29.00,-30.00,-31.00,-32.00); //c
                                                                                                    //    -4,    -3,    -2,    -1,     0,     1,     2,     3,     4,     5,     6,     7,     8,     9,    10
  QUALITY_CLIPS_PER_CHANNEL             : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Integer  = (     3,     3,     3,     3,     3,     3,     3,     3,     2,     1,     0,     0,     0,     0,     0);
  //==============================================================================
var
  Unit_Data_Packer2 : Array[0..14] of Integer;

  //============================================================================
  // Check command line parameters from nparameter;
  //============================================================================
  procedure check_switches;
  //============================================================================
  begin
    settings.quality_double:=quality;
    settings.quality_integer:=floor(settings.quality_double);
    settings.quality_fraction:=settings.quality_double-settings.quality_integer;

    if analyses<>2 then
      parameter_string:=parameter_string+' --analyses '+inttostr(analyses);

    if impulse then
      parameter_string:=parameter_string + ' --impulse';

    if maxclips()>-1 then
      settings.maximum_clips_per_channel:=maxclips()
    else
      settings.maximum_clips_per_channel:=QUALITY_CLIPS_PER_CHANNEL[min(QUALITY_PRESET_MAX,settings.quality_integer+ord(settings.quality_fraction<>0))];

    if scaling<>-1 then
    begin
      settings.scaling_factor:=scaling;
      parameter_string:=parameter_string+' --scale '+floattostrf(settings.scaling_factor,fffixed,99,6);
    end
    else
      settings.scaling_factor:=1;

    if settings.scaling_factor=0 then
      lossyWAVerror('Scale value must not be zero',$21);
    settings.scaling_factor_inv:=1/settings.scaling_factor;

    if linkchannels or midside then
      parameter_string:=parameter_string+' --linkchannels';

    if shaping>-1 then
    begin
      settings.shaping_factor:=shaping;
      parameter_string:=parameter_string+' --shaping '+floattostrf(settings.shaping_factor,ffFixed,99,4);
    end
    else
      if shaping=-1 then
        settings.shaping_factor:=0
      else
        if altpreset then
          settings.shaping_factor:=min(1,(power(max(settings.quality_double/10,0),2.58496250072115)+max(settings.quality_double/10,0))*single_half)
        else
          settings.shaping_factor:=min(1,max(0,settings.quality_double)*OneOver[10]);

    settings.shaping_is_on:=(settings.shaping_factor>0);

    if midside() then
      parameter_string:=parameter_string+' --midside';

    if limit<>-1 then
    begin
      parameter_string:=parameter_string + ' --limit '+floattostrf(limit,ffFixed,99,0);
      Global.Upper_Freq_Limit:=limit;
    end
    else
      if altpreset then
        Global.Upper_Freq_Limit:=UPPER_FREQ_LIMIT_ALT
      else
        Global.Upper_Freq_Limit:=UPPER_FREQ_LIMIT;

    Global.Lower_Freq_Limit:=LOWER_FREQ_LIMIT;

    if underlap>0 then
      parameter_string:=parameter_string+' --underlap '+inttostr(underlap);

    if sortspread then
      parameter_string:=parameter_string + ' --sortspread';

    if altpreset then
      parameter_string:=parameter_string + ' --altpreset';


    if altpreset then
    begin
      if Settings.quality_integer=QUALITY_PRESET_MAX then
      begin
        settings.noise_threshold_shift_minimum:=QUALITY_NOISE_THRESHOLD_SHIFTS_ALT[QUALITY_PRESET_MAX];
        settings.noise_threshold_shift_average:=QUALITY_SIGNAL_TO_NOISE_RATIOS_ALT[QUALITY_PRESET_MAX];
        settings.dynamic_minimum_bits_to_keep:=QUALITY_MINIMUM_BITS_TO_KEEP_ALT[QUALITY_PRESET_MAX];
      end
      else
      begin
        settings.noise_threshold_shift_minimum:=Settings.quality_fraction*(QUALITY_NOISE_THRESHOLD_SHIFTS_ALT[Settings.quality_integer+1]-QUALITY_NOISE_THRESHOLD_SHIFTS_ALT[Settings.quality_integer])+QUALITY_NOISE_THRESHOLD_SHIFTS_ALT[Settings.quality_integer];
        settings.noise_threshold_shift_average:=Settings.quality_fraction*(QUALITY_SIGNAL_TO_NOISE_RATIOS_ALT[Settings.quality_integer+1]-QUALITY_SIGNAL_TO_NOISE_RATIOS_ALT[Settings.quality_integer])+QUALITY_SIGNAL_TO_NOISE_RATIOS_ALT[Settings.quality_integer];
        settings.dynamic_minimum_bits_to_keep:=Settings.quality_fraction*(QUALITY_MINIMUM_BITS_TO_KEEP_ALT[Settings.quality_integer+1]-QUALITY_MINIMUM_BITS_TO_KEEP_ALT[Settings.quality_integer])+QUALITY_MINIMUM_BITS_TO_KEEP_ALT[Settings.quality_integer];
      end;
      settings.noise_threshold_shift_minimum:=round(settings.noise_threshold_shift_minimum*1000)*0.001;
      settings.noise_threshold_shift_average:=round(settings.noise_threshold_shift_average*1000)*0.001;
      settings.dynamic_minimum_bits_to_keep:=round(settings.dynamic_minimum_bits_to_keep*1000)*0.001;
    end
    else
    begin
      if Settings.quality_integer=QUALITY_PRESET_MAX then
      begin
        settings.noise_threshold_shift_minimum:=QUALITY_NOISE_THRESHOLD_SHIFTS[QUALITY_PRESET_MAX];
        settings.noise_threshold_shift_average:=QUALITY_SIGNAL_TO_NOISE_RATIOS[QUALITY_PRESET_MAX];
        settings.dynamic_minimum_bits_to_keep:=QUALITY_MINIMUM_BITS_TO_KEEP[QUALITY_PRESET_MAX];
      end
      else
      begin
        settings.noise_threshold_shift_minimum:=Settings.quality_fraction*(QUALITY_NOISE_THRESHOLD_SHIFTS[Settings.quality_integer+1]-QUALITY_NOISE_THRESHOLD_SHIFTS[Settings.quality_integer])+QUALITY_NOISE_THRESHOLD_SHIFTS[Settings.quality_integer];
        settings.noise_threshold_shift_average:=Settings.quality_fraction*(QUALITY_SIGNAL_TO_NOISE_RATIOS[Settings.quality_integer+1]-QUALITY_SIGNAL_TO_NOISE_RATIOS[Settings.quality_integer])+QUALITY_SIGNAL_TO_NOISE_RATIOS[Settings.quality_integer];
        settings.dynamic_minimum_bits_to_keep:=Settings.quality_fraction*(QUALITY_MINIMUM_BITS_TO_KEEP[Settings.quality_integer+1]-QUALITY_MINIMUM_BITS_TO_KEEP[Settings.quality_integer])+QUALITY_MINIMUM_BITS_TO_KEEP[Settings.quality_integer];
      end;
      settings.noise_threshold_shift_minimum:=round(settings.noise_threshold_shift_minimum*1000)*0.001;
      settings.noise_threshold_shift_average:=round(settings.noise_threshold_shift_average*1000)*0.001;
      settings.dynamic_minimum_bits_to_keep:=round(settings.dynamic_minimum_bits_to_keep*1000)*0.001;
    end;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  // Setup most of the internal arrays of values to be used in the analyses;
  //============================================================================
  procedure initial_setup;
  //============================================================================
  var
    sa_i, sa_j : Integer;
    this_channel                          : Integer;
    this_analysis_number                  : Integer;
    total_overlap_length                  : Integer;
    quality_temp                          : Double;
  begin
    settings.fft_bit_length_shift:=Global.codec_block_bits-9;
    settings.fft_lengths_shift:=power(2,settings.fft_bit_length_shift);

    settings.fft_analysis_switches:=floor(STANDARD_FFT_LENGTHS*settings.fft_lengths_shift);

    if analyses<>2 then
    begin
      if analyses>2 then
        settings.fft_analysis_switches:=settings.fft_analysis_switches or floor(ADDED_FFT_LENGTH_A*settings.fft_lengths_shift);

      if analyses>3 then
        settings.fft_analysis_switches:=settings.fft_analysis_switches or floor(ADDED_FFT_LENGTH_B*settings.fft_lengths_shift);

      if analyses>4 then
        settings.fft_analysis_switches:=settings.fft_analysis_switches or floor(ADDED_FFT_LENGTH_C*settings.fft_lengths_shift);
    end;

    if (Settings.quality_integer>2) or impulse then
      settings.fft_analysis_switches:=settings.fft_analysis_switches or floor(IMPULSE_FFT_LENGTH*settings.fft_lengths_shift);

    if Global.Upper_Freq_Limit>samplerate*single_half then
      Global.Upper_Freq_Limit:=floor(samplerate*single_half);

    if (not silent) then
    begin
      if (verbosity) then
      begin
        Write(ErrOutput,'Filename  : ',WAVFilePrintName,CRLF,'File Info : ',samplerate*0.001:0:2,'kHz; ',Global.channels,' channel; ',bitspersample,' bit; ');

        if wavesize<MAX_WAVE_SIZE then
        begin
          time_string_make(strings.time,totalsamples/samplerate);
          size_string_make(strings.size,totalsamples*Global.channels*bytespersample);
          Write(ErrOutput,strings.time,', ',strings.size,' raw PCM');
        end;

        Write(ErrOutput,CRLF+'Progress  : ');
      end
      else
        Write(ErrOutput,WAVFilePrintName,';');
    end;
    flush(ErrOutput);

    for this_channel:=0 to Global.channels-1 do
    begin
      Global.channel:=this_channel;
      for this_analysis_number:=1 to PRECALC_ANALYSES do
        results.saved_fft_results[Global.channel,this_analysis_number].start:=-1;
    end;

    Global.clipped_samples:=0;

    Remove_Bits_Init;

    //==========================================================================
    // calculating for each fft_bit_length
    //==========================================================================
    for sa_i:=1 to MAX_FFT_BIT_LENGTH do
    begin
      Set_Global_FFT_Variables(sa_i);

      if FFTW_Initialised then
        if (((settings.fft_analysis_switches and PowersOf.twoint[sa_i])>0)) then
          fftw_plans[Global.fft_bit_length]:=fftw_plan_dft_r2c_1d(Global.fft_length,@FFT_Array,@FFT_Array,$69);

      // Calculate actual analysis_time values for fft_lengths
      process.bin_width[Global.fft_bit_length]:=samplerate*Global.fft_length_recip;

      // Calculate which FFT bins correspond to the frequency limits
      Bins.Lo[Global.fft_bit_length]:=max(1,round(Global.Lower_Freq_Limit/process.bin_width[Global.fft_bit_length])-1);
      Bins.Hi[Global.fft_bit_length]:=max(1,round(Global.Upper_Freq_Limit/process.bin_width[Global.fft_bit_length])-1);
      Bins.Max_Recip[Global.fft_bit_length]:=1/(Bins.Hi[Global.fft_bit_length]-Bins.Lo[Global.fft_bit_length]+1);

      process.end_overlap_length[Global.fft_bit_length]:=min(Global.fft_length div 2,Global.codec_block_size);
      process.actual_analysis_blocks_start[Global.fft_bit_length]:=-process.end_overlap_length[Global.fft_bit_length];

      total_overlap_length:=Global.codec_block_size+process.end_overlap_length[Global.fft_bit_length]*2-Global.fft_length;
      process.fft_underlap_length[Global.fft_bit_length]:=max(1,Global.fft_length div max(2,underlap));

      process.analysis_blocks[Global.fft_bit_length]:=max(0,round(total_overlap_length/process.fft_underlap_length[Global.fft_bit_length]));
      process.fft_underlap_length[Global.fft_bit_length]:=total_overlap_length/max(1,process.analysis_blocks[Global.fft_bit_length]);
    end;

    //==========================================================================
    // calculating for each analysis_number
    //==========================================================================
    for this_analysis_number:=1 to PRECALC_ANALYSES do
    begin
      Global.analysis_number:=this_analysis_number;
      Set_Global_FFT_Variables(global.codec_block_bits-5+Global.analysis_number);

      Bins.Vl[Global.fft_bit_length]:=max(1,round(VAR_LO_FREQ[Global.analysis_number]/process.bin_width[Global.fft_bit_length])-1);

      // Calculate which FFT bin corresponds to the low frequency limit
      for sa_i:=0 to SPREAD_ZONES+1 do
      begin
        frequency_bins[Global.analysis_number,sa_i]:=min(Global.fft_length div 2,max(1,round(FREQUENCY_LIMITS[sa_i]/process.bin_width[Global.fft_bit_length])))-1;
        if frequency_bins[Global.analysis_number,sa_i]>Global.fft_length div 2-1 then
          lossyWAVerror('frequency too high : '+floattostrf(FREQUENCY_LIMITS[sa_i]/1000,ffFixed,99,3)+'kHz',$21);
      end;

      for sa_j:=SPREAD_ZONES downto 0 do
        for sa_i:=frequency_bins[Global.analysis_number,sa_j] to frequency_bins[Global.analysis_number,sa_j+1] do
        {$R-}
          spreading_averages_ptr[Global.analysis_number]^[sa_i].width:=min(max(2,Bins.Hi[Global.fft_bit_length]-sa_i+1),SPREADING_FUNCTION_ARRAY[Global.analysis_number,sa_j]);
        {$R+}

      for sa_i:=0 to Bins.Vl[Global.fft_bit_length] do
      {$R-}
        spreading_averages_ptr[Global.analysis_number]^[sa_i].fraction:=0;
      {$R+}

      for sa_i:=Bins.Vl[Global.fft_bit_length] to Bins.Hi[Global.fft_bit_length] do
      {$R-}
        spreading_averages_ptr[Global.analysis_number]^[sa_i].fraction:=floor(SPREADING_STEPS*max(0,min((UPPER_FREQ_LIMIT-VAR_LO_FREQ[Global.analysis_number]),sa_i*process.bin_width[Global.fft_bit_length]-VAR_LO_FREQ[Global.analysis_number]))/(UPPER_FREQ_LIMIT-VAR_LO_FREQ[Global.analysis_number])*spreading_function_widths[Global.analysis_number]);
      {$R+}

      Bins.Max[Global.fft_bit_length]:=Bins.Hi[Global.fft_bit_length]-1;

      Bins.Old_Recip[Global.fft_bit_length]:=1/(Bins.Max[Global.fft_bit_length]-Bins.Lo[Global.fft_bit_length]+1);
    //==========================================================================
    end; // calculating for each analysis_number
    //==========================================================================

    Global.output_blocks:=floor(PowersOf.two[20-Global.codec_block_bits]*OneOver[Global.channels]);
    Global.last_print:=0;

    Global.blocks_processed:=1;
    Global.blocks_processed_recip:=1;

    Global.bits_removed:=0;
    Global.bits_lost:=0;
  //============================================================================
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('nInitialise',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
