//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nOutput;
//==============================================================================
interface
//==============================================================================
{$DEFINE SAMPLEDIST_ON}
//==============================================================================
type
  bit_removal_history_array = array[0..4194303] of byte;
var
  Unit_Data_Packer1   : Array[0..15] of Integer;
  bit_removal_history : ^bit_removal_history_array;
  corr_nonzero_count  : Cardinal = 0;
//==============================================================================
procedure remove_bits_detailed_output(var ToOutput : Text);
Procedure Write_Results(Var ToOutput : Text);
procedure write_cleanup;
{$IFDEF SAMPLEDIST_ON}
procedure lsb_analysis;
Procedure Write_blockdist(var ToOutput : Text);
Procedure Write_SampleDist(var ToOutput : Text);
{$ENDIF}
Procedure nOutput_Init;
Procedure Process_Output;
//==============================================================================
implementation
//==============================================================================
uses
  fftw_interface,
  nParameter,
  nSystem,
  nWav,
  Math,
  SysUtils;
//==============================================================================
const
  hyphen_string                         : String[31] = '------------------------------';
  bits_filled                           : String[31] = '##############################';
  bits_empty                            : String[31] = '..............................';
var
  wave_orvalue                          : Integer;
  btrd_orvalue                          : Integer;

  wave_block_lsb                        : Array[0..MAX_CHANNELS-1,0..33] of integer;
  btrd_block_lsb                        : Array[0..MAX_CHANNELS-1,0..33] of integer;
  wave_block_msb                        : Array[0..MAX_CHANNELS-1,0..33] of integer;
  btrd_block_msb                        : Array[0..MAX_CHANNELS-1,0..33] of integer;
  wave_sample_lsb                       : Array[0..MAX_CHANNELS-1,0..33] of integer;
  btrd_sample_lsb                       : Array[0..MAX_CHANNELS-1,0..33] of integer;
  wave_sample_msb                       : Array[0..MAX_CHANNELS-1,0..33] of integer;
  btrd_sample_msb                       : Array[0..MAX_CHANNELS-1,0..33] of integer;

  Unit_Data_Packer2                     : Array[0..7] of Integer;
//==============================================================================

  //============================================================================
  procedure remove_bits_detailed_output(var ToOutput : Text);
  //============================================================================
  const
    time_width = 10;
    summary_width = 4;
  var
    rb_i, rb_j, rb_k, rb_l : Integer;
    ss_i, ss_j, ss_k : Integer;
    btr_row_tot : Double;
    sep_string : String;
    channel : Integer;
    time_string : String;
  Begin
    ss_j:=Global.channels*3;
    ss_i:=(80-time_width-summary_width-1) div ss_j;
    ss_k:=ss_i*Global.channels;

    sep_string:='+--------+';
    for rb_i:=0 to ss_i-1 do
      sep_string:=sep_string+copy(hyphen_string,1,ss_j-1)+'+';
    sep_string:=sep_string+'---+';
    write(ToOutput,sep_string,CRLF,'|  Time  |');
    for rb_i:=0 to ss_i-1 do
      for channel:=0 to Global.channels-1 do
      begin
        write(ToOutput,'#',channel);
        if channel=Global.channels-1 then
          write(ToOutput,'|')
        else
          write(ToOutput,',');
      end;
    write(ToOutput,'Tot|',CRLF,sep_string,CRLF);
    btr_row_tot:=0;
    rb_j:=0;
    for rb_i:=0 to Global.blocks_processed-1 do
      for channel:=0 to Global.channels-1 do
      begin
        rb_l:=rb_j mod ss_k;
        if rb_l=0 then
        begin
          time_string_make(time_string,rb_i*Global.codec_block_size/Global.sample_rate);
          btr_row_tot:=0;
          write(ToOutput,'|',time_string,'|');
        end;
        rb_k:=bit_removal_history^[rb_j];
        btr_row_tot:=btr_row_tot+rb_k;
        write(ToOutput,rb_k:2);
        if channel=Global.channels-1 then
          write(ToOutput,'|')
        else
          write(ToOutput,',');

        if rb_l=ss_k-1 then
        begin
          write(ToOutput,btr_row_tot:3:0,'|',CRLF);
        end;
        inc(rb_j);
      end;

      rb_l:=rb_j mod ss_k;
      if (rb_l<ss_k-1) and (rb_l>0) then
      begin
        for rb_i:=0 to (ss_k-rb_l-1) div Global.channels do
          for channel:=0 to Global.channels-1 do
            if channel<Global.channels-1 then
              write(ToOutput,'   ')
            else
              write(ToOutput,'  |');
        write(ToOutput,btr_row_tot:3:0,'|',CRLF);
      end;
    Write(ToOutput,sep_string,CRLF);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Write_Results(Var ToOutput : Text);
  //============================================================================
  begin
    Write(ToOutput,'Results   : ',
                    OneOver[Global.channels]*Global.bits_removed*Global.blocks_processed_recip:0:4,' bits; ',Global.processing_rate:0:2,'x; ',strings.elapsed);

    {$IFDEF SAMPLEDIST_ON}
    if blockdist() or sampledist() then
      Write(ToOutput,'; ',corr_nonzero_count/(Global.samples_processed*Global.channels)*100:0:2,'% modified.'); //Overall samples changed;
    {$ENDIF}

    write(ToOutput,CRLF);

    if Global.bits_lost+Global.clipped_samples>0 then
    begin
      Write(ToOutput,'Clipping  :');
      if Global.bits_lost*OneOver[Global.channels]*Global.blocks_processed_recip>0.00005 then
        Write(ToOutput,Global.bits_lost*OneOver[Global.channels]*Global.blocks_processed_recip:7:4,' bits lost,');
      Write(ToOutput,' ',Global.clipped_samples,' clip');
      if Global.clipped_samples<>1 then
        Write(ToOutput,'s');
      Write(ToOutput,'.'+CRLF);
    end;

    {$IFDEF SAMPLEDIST_ON}
    if blockdist() then
      write_blockdist(ToOutput);

    if sampledist() then
      write_sampledist(ToOutput);
    {$ENDIF}

    if detailedoutput() then
      remove_bits_detailed_output(ToOutput);

    Flush(ToOutput);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure write_cleanup;
  //============================================================================
  var
    wc_i : Integer;
  begin
    if FFTW_Initialised then
    Begin
      for wc_i:=1 to MAX_FFT_BIT_LENGTH do
        if fftw_plans[wc_i]<>nil then
          FFTW_DESTROY_PLAN(fftw_plans[wc_i]);

      FFTW_Cleanup();
    End;

    GetTimer();
    time_string_make(strings.elapsed,timer.elapsed);

    if (not silent) then
    begin
      if (verbosity) then
      begin
        Write(ErrOutput,#13+'                                                                               '+#13);
        Write_Results(ErrOutput);
      end
      else
        write(ErrOutput,OneOver[Global.channels]*Global.bits_removed*Global.blocks_processed_recip:7:4,';',Global.processing_rate:6:2,'x; ',strings.elapsed,CRLF);
    end;

    if (writetolog) then
    begin
      time_string_make(strings.time,Global.samples_processed/samplerate);
      size_string_make(strings.size,1.0*Global.samples_processed*Global.channels*bytespersample);

      open_log_file;

      Write(LogOutput,lossyWAVhead1,
                      'Processed : ',datestamp_String,CRLF+
                      'Settings  : ',parameter_string,CRLF+
                      'Filename  : ',WavFilePrintName,CRLF+
                      'File Info : ',samplerate/1000:0:2,'kHz; ',Global.channels,' channel; ',bitspersample,' bit, ',strings.time,', ',strings.size,CRLF);
      Write_Results(LogOutput);
      Write(LogOutput,CRLF);

      close_log_file;
    end;
  //============================================================================
  end;
  //============================================================================

{$IFDEF SAMPLEDIST_ON}
  //============================================================================
  procedure lsb_analysis;
  //============================================================================
  asm
                push    ebp
                push    ebx
                push    edi
                push    esi

                mov     esi,Global.channels

      @@04:     sub     esi,1
                jl      @@99

                mov     eax,esi
                mov     eax,dword ptr CHANNEL_OFFSET[eax*POINTER_SIZE]
                mov     edx,esi
                shl     edx,8

                mov     ecx,this_codec_block_size
                mov     wave_orvalue,0
                mov     btrd_orvalue,0

                mov     ebp,dword ptr rotating_blocks_ptr[2*POINTER_SIZE]
                add     ebp,eax

      @@05:     sub     ecx,1
                jl      @@10

                mov     ebx,dword ptr [ebp+ecx*INTEGER_SIZE]

                bsf     edi,ebx
                jnz     @@06

                add     dword ptr wave_sample_lsb[edx+0*INTEGER_SIZE],1
                add     dword ptr wave_sample_msb[edx+0*INTEGER_SIZE],1
                jmp     @@07

      @@06:     add     dword ptr wave_sample_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                or      ebx,ebx
                jns     @@06a

                neg     ebx

      @@06a:    bsr     edi,ebx
                add     dword ptr wave_sample_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                or      dword ptr wave_orvalue,ebx

      @@07:     mov     ebx,dword ptr btrd_codec_block[eax+ecx*INTEGER_SIZE]

                bsf     edi,ebx
                jnz     @@08

                add     dword ptr btrd_sample_lsb[edx+0*INTEGER_SIZE],1
                add     dword ptr btrd_sample_msb[edx+0*INTEGER_SIZE],1
                jmp     @@09

      @@08:     add     dword ptr btrd_sample_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                or      ebx,ebx
                jns     @@08a

                neg     ebx

      @@08a:    bsr     edi,ebx
                add     dword ptr btrd_sample_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                or      dword ptr btrd_orvalue,ebx

      @@09:     sub     ebx,ebx
                add     ebx,dword ptr corr_codec_block[eax+ecx*INTEGER_SIZE]
                jz      @@09a

                add     corr_nonzero_count,1

      @@09a:    jmp     @@05

      @@10:     bsf     edi,dword ptr wave_orvalue
                jnz     @@11

                add     dword ptr wave_block_lsb[edx+0],1
                add     dword ptr wave_block_msb[edx+0],1
                jmp     @@12

      @@11:     add     dword ptr wave_block_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                bsr     edi,dword ptr wave_orvalue
                add     dword ptr wave_block_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

      @@12:     bsf     edi,dword ptr btrd_orvalue
                jnz     @@13

                add     dword ptr btrd_block_lsb[edx+0],1
                add     dword ptr btrd_block_msb[edx+0],1
                jmp     @@14

      @@13:     add     dword ptr btrd_block_lsb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1
                bsr     edi,dword ptr btrd_orvalue
                add     dword ptr btrd_block_msb[edx+edi*INTEGER_SIZE+INTEGER_SIZE],1

      @@14:     jmp     @@04

      @@99:     pop     esi
                pop     edi
                pop     ebx
                pop     ebp
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure Write_blockdist(var ToOutput : Text);
  //============================================================================
  var
    nd_i, nd_j : Integer;
    this_wave_lsb_percentage              : single;
    this_btrd_lsb_percentage              : single;
    this_wave_msb_percentage              : single;
    this_btrd_msb_percentage              : single;
    bits_filled_wave_lsb                  : Integer;
    bits_filled_btrd_lsb                  : Integer;
    bits_filled_wave_msb                  : Integer;
    bits_filled_btrd_msb                  : Integer;
  //============================================================================
  begin
    Write(ToOutput,CRLF+'Codec-block lsb / msb distribution.'+CRLF+
                        '+---+------------------------------------+------------------------------------+'+CRLF+
                        '|Bit|Input codec-block lsb distribution  |Input codec-block msb distribution  |'+CRLF+
                        '+---+-------+----------------------------+-------+----------------------------+'+CRLF);

    for nd_i:=0 to bitspersample do
    begin
      this_wave_lsb_percentage:=0;
      this_wave_msb_percentage:=0;

      for nd_j:=0 to Global.channels-1 do
      begin
        this_wave_lsb_percentage:=this_wave_lsb_percentage+wave_block_lsb[nd_j,nd_i]*Global.blocks_processed_recip;
        this_wave_msb_percentage:=this_wave_msb_percentage+wave_block_msb[nd_j,nd_i]*Global.blocks_processed_recip;
      end;
      bits_filled_wave_lsb:=round(this_wave_lsb_percentage*28);
      bits_filled_wave_msb:=round(this_wave_msb_percentage*28);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_wave_lsb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave_lsb),copy(bits_empty,1,28-bits_filled_wave_lsb),'|');
      Write(ToOutput,floattostrf(this_wave_msb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave_msb),copy(bits_empty,1,28-bits_filled_wave_msb),'|',CRLF);
    end;

    Write(ToOutput,'+---+------------------------------------+------------------------------------+'+CRLF+
                   '|Bit|Output codec-block lsb distribution |Output codec-block msb distribution |'+CRLF+
                   '+---+-------+----------------------------+-------+----------------------------+'+CRLF);

    for nd_i:=0 to bitspersample do
    begin
      this_btrd_lsb_percentage:=0;
      this_btrd_msb_percentage:=0;
      for nd_j:=0 to Global.channels-1 do
      begin
        this_btrd_lsb_percentage:=this_btrd_lsb_percentage+btrd_block_lsb[nd_j,nd_i]*Global.blocks_processed_recip;
        this_btrd_msb_percentage:=this_btrd_msb_percentage+btrd_block_msb[nd_j,nd_i]*Global.blocks_processed_recip;
      end;
      bits_filled_btrd_lsb:=round(this_btrd_lsb_percentage*28);
      bits_filled_btrd_msb:=round(this_btrd_msb_percentage*28);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_btrd_lsb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd_lsb),copy(bits_empty,1,28-bits_filled_btrd_lsb),'|');
      Write(ToOutput,floattostrf(this_btrd_msb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd_msb),copy(bits_empty,1,28-bits_filled_btrd_msb),'|',CRLF);
    end;

    Write(ToOutput,'+---+-------+----------------------------+-------+----------------------------+',CRLF);
  end;
  //============================================================================

  //============================================================================
  Procedure Write_SampleDist(var ToOutput : Text);
  //============================================================================
  var
    nd_i, nd_j : Integer;
    this_wave_lsb_percentage              : single;
    this_btrd_lsb_percentage              : single;
    this_wave_msb_percentage              : single;
    this_btrd_msb_percentage              : single;
    bits_filled_wave_lsb                  : Integer;
    bits_filled_btrd_lsb                  : Integer;
    bits_filled_wave_msb                  : Integer;
    bits_filled_btrd_msb                  : Integer;
  //============================================================================
  begin
    Write(ToOutput,CRLF+'Sample lsb / msb distribution.'+CRLF+
                        '+---+------------------------------------+------------------------------------+'+CRLF+
                        '|Bit|Input sample lsb distribution       |Input sample msb distribution       |'+CRLF+
                        '+---+-------+----------------------------+-------+----------------------------+'+CRLF);

    for nd_i:=0 to bitspersample do
    begin
      this_wave_lsb_percentage:=0;
      this_wave_msb_percentage:=0;
      for nd_j:=0 to Global.channels-1 do
      begin
        this_wave_lsb_percentage:=this_wave_lsb_percentage+wave_sample_lsb[nd_j,nd_i]/Global.samples_processed;
        this_wave_msb_percentage:=this_wave_msb_percentage+wave_sample_msb[nd_j,nd_i]/Global.samples_processed;
      end;
      bits_filled_wave_lsb:=round(this_wave_lsb_percentage*28);
      bits_filled_wave_msb:=round(this_wave_msb_percentage*28);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_wave_lsb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave_lsb),copy(bits_empty,1,28-bits_filled_wave_lsb),'|');
      Write(ToOutput,floattostrf(this_wave_msb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_wave_msb),copy(bits_empty,1,28-bits_filled_wave_msb),'|',CRLF);
    end;

    Write(ToOutput,'+---+-------+----------------------------+-------+----------------------------+'+CRLF+
                   '|Bit|Output sample lsb distribution      |Output sample msb distribution      |'+CRLF+
                   '+---+-------+----------------------------+-------+----------------------------+'+CRLF);

    for nd_i:=0 to bitspersample do
    begin
      this_btrd_lsb_percentage:=0;
      this_btrd_msb_percentage:=0;
      for nd_j:=0 to Global.channels-1 do
      begin
        this_btrd_lsb_percentage:=this_btrd_lsb_percentage+btrd_sample_lsb[nd_j,nd_i]/Global.samples_processed;
        this_btrd_msb_percentage:=this_btrd_msb_percentage+btrd_sample_msb[nd_j,nd_i]/Global.samples_processed;
      end;
      bits_filled_btrd_lsb:=round(this_btrd_lsb_percentage*28);
      bits_filled_btrd_msb:=round(this_btrd_msb_percentage*28);

      if nd_i=0 then
        Write(ToOutput,'|NUL')
      else
        Write(ToOutput,'|',nd_i-1:3);

      Write(ToOutput,'|',floattostrf(this_btrd_lsb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd_lsb),copy(bits_empty,1,28-bits_filled_btrd_lsb),'|');
      Write(ToOutput,floattostrf(this_btrd_msb_percentage*100,fffixed,99,2):6,'%|');
      Write(ToOutput,copy(bits_filled,1,bits_filled_btrd_msb),copy(bits_empty,1,28-bits_filled_btrd_msb),'|',CRLF);
    end;

    Write(ToOutput,'+---+-------+----------------------------+-------+----------------------------+',CRLF);
  //============================================================================
  end;
  //============================================================================
{$ENDIF}

  //============================================================================
  Procedure Process_Output;
  //============================================================================
  Begin
    if (verbosity) and (not silent) and ((Global.last_print=0) or (Global.blocks_processed mod Global.Output_Blocks=0)) then
    begin
      GetTimer;
      time_string_make(strings.elapsed,timer.elapsed);

      Write(ErrOutput,#13,'Progress  :');

      if wavesize()=MAX_WAVE_SIZE then
      begin
        size_string_make(strings.size,1.0*Global.samples_processed*Global.channels*bytespersample);
        Write(ErrOutput,' ',strings.size,';')
      end
      else
        Write(ErrOutput,Global.samples_processed/totalsamples*100:6:2,'%;');

      write(ErrOutput,OneOver[Global.channels]*Global.bits_removed*Global.blocks_processed_recip:7:4,' bits;',Global.processing_rate:6:2,'x;');

      {$IFDEF SAMPLEDIST_ON}
      if blockdist() or sampledist() then
        Write(ErrOutput,corr_nonzero_count*OneOver[Global.channels]/Global.samples_processed*100:6:2,'%;');
      {$ENDIF}

      Write(ErrOutput,' ',strings.elapsed);
      if (Global.samples_processed>0) then
      begin
        time_string_make(strings.estimate,min(86399,totalsamples/Global.samples_processed*timer.elapsed));
        Write(ErrOutput,'/',strings.estimate,' ');
      end;

      Global.last_print:=Global.blocks_processed;
    end;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure nOutput_Init;
  //============================================================================
  Begin
    if detailedoutput then
      new(bit_removal_history);
  End;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('nOutput',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}
//==============================================================================

//==============================================================================
finalization
//==============================================================================
  if bit_removal_history<>nil then
    dispose(bit_removal_history);
//==============================================================================

//==============================================================================
end.
//==============================================================================
