//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nParameter;
//==============================================================================
interface
//==============================================================================
Uses
  nSystem;
//==============================================================================
var
  Unit_Data_Packer1 : Array[0..15] of Integer;
//==============================================================================
Const
  version_string='lossyWAV 1.2.0b';
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAVHead1=version_string+', Copyright (C) 2007,2008,2009 Nick Currie. Copyleft.'+#13#10;
  lossyWAVhead2= 'This is free software under the GNU GPLv3+ license; There is NO WARRANTY, to'+CRLF+
                 'the extent permitted by law. <http://www.gnu.org/licenses/> for details.'+CRLF;
//==============================================================================
var
  Parameter_String : string;
//==============================================================================
procedure lossyWAVerror  (lwe_string : string; lwe_value : integer);
procedure lossyWAVwarning(lww_string : string);

function  DirInp                : string;   // name of the source file folder.
function  DirOut                : String;   // name of the output file folder.

function  wavInp                : string;   // name of the input wav file & directory
function  wavOut                : string;   // name of the input wav file & directory

function  lossyInp              : String;   // name of the lossy input wav file.
function  lossyOut              : string;   // name of the lossy output wav file based on the input file name.

function  lwcdfInp              : String;   // name of the lwcdf input wav file.
function  lwcdfOut              : string;   // name of the lwcdf output wav file based on the input file name.

Function  WAVFilePrintName      : String;   // name of WAV file OR 'from STDIN' if STDINPUT=1.
function  lossyFileName         : string;   // output filename only;
function  lwcdfFileName         : string;   // correction file output filename only;
function  logfilename           : String;   // name of lossyWAV log file.
function  outputname            : String;   // output filename if STDIN input selected.

function  quality               : Double;   // quality level. default=2;
function  shaping               : Double;   // enable optional fixed noise shaping; default=off.
function  limit                 : Integer;  // modify upper bound of frequency range analysed.

function  scaling               : Double;   // scaling factor from WaveGain;
function  priority              : Integer;  // below or low_priority process priority switch;
function  analyses              : Integer;  // number of FFT analysis lengths, default=2;
function  maxclips              : Integer;  // set maximum number of clips permissible per channel per codec-block.

function  forcing               : longbool; // forcibly over-write output file if it exists.
function  correct               : longbool; // write correction file if correct=1.
function  checking              : longbool; //
function  merging               : longbool; //

function  silent                : longbool; // enable silent running.
function  verbosity             : longbool; // verbosity level. default=false;
function  warnings              : longbool; // warnings switch state. default=false;
function  writetolog            : longbool; // enable writing to lossyWAV.log file; default=0.

function  detailedoutput        : longbool; // debug switch state. default=false
function  linkchannels          : longbool; // revert to "old" method of one bits-to-remove value for all channels.
function  blockdist             : longbool; // show bit distribution of WAVE / BTRD output codec blocks;
function  sampledist            : longbool; // show bit distribution of WAVE / BTRD output samples;

function  STDINPUT              : longbool; // read from STDIN;
function  STDOUTPUT             : longbool; // write to STDOUTPUT;
function  impulse               : longbool; // enable 32 sample FFT analysis for enhanced impulse detection.
function  midside               : longbool; // enable mid/side channel analysis for 2 channel audio.

function  underlap              : integer;  // select experimental underlap mode length.
function  sortspread            : longbool; // enable experimental spreading function.
function  altpreset             : longbool;

procedure open_log_file;
procedure close_log_file;

//==============================================================================
implementation
//==============================================================================
uses
  nmaths,
  nwav,
  SysUtils,
  Windows,
  Math;
//==============================================================================
//nparameter constants, variables & types
//==============================================================================
const
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
    lossyWAV_GPL =
                 'This program is free software: you can redistribute it and/or modify it under'+
            CRLF+'the terms of the GNU General Public License as published by the Free Software'+
            CRLF+'Foundation, either version 3 of the License, or (at your option) any later'+
            CRLF+'version.'+CRLF+
            CRLF+'This program is distributed in the hope that it will be useful,but WITHOUT ANY'+
            CRLF+'WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A'+
            CRLF+'PARTICULAR PURPOSE.  See the GNU General Public License for more details.'+CRLF+
            CRLF+'You should have received a copy of the GNU General Public License along with'+
            CRLF+'this program.  If not, see <http://www.gnu.org/licenses/>.'+CRLF;
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
    lossyWAV_process_description =
                 'Process Description:'+CRLF+
            CRLF+'lossyWAV is a near lossless audio processor which dynamically reduces the '+
            CRLF+'bitdepth of the signal on a block-by-block basis. Bitdepth reduction adds white'+
            CRLF+'noise to the processed output. The amount of permissible added noise is based'+
            CRLF+'on analysis of the signal levels in the default frequency range 20Hz to 16kHz.'+CRLF+
            CRLF+'If signals above the upper limiting frequency are at an even lower level, they'+
            CRLF+'can be swamped by the added noise. This is usually inaudible, but the behaviour'+
            CRLF+'can be changed by specifying a different --limit (in the range 10kHz to 20kHz).'+CRLF+
            CRLF+'For many audio signals there is little content at very high frequencies and'+
            CRLF+'forcing lossyWAV to keep the added noise level lower than the content at these'+
            CRLF+'frequencies can increase the bitrate dramatically for no perceptible benefit.'+CRLF;
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
    lossyWAV_standard_help =
                 'Usage   : lossyWAV <input wav file> <options>'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Example : lossyWAV musicfile.wav'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Quality Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-I, --insane         highest quality output, suitable for transcoding;'+
            CRLF+'-E, --extreme        high quality output, also suitable for transcoding;'+
            CRLF+'-S, --standard       default quality output, considered to be transparent;'+
            CRLF+'-P, --portable       good quality output for DAP use, not fully transparent.'+
            CRLF+'-Z, --zero           lowest quality preset, probably contains artifacts.'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'Standard Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-C, --correction     write correction file for processed WAV file; default=off.'+
            CRLF+'-f, --force          forcibly over-write output file if it exists; default=off.'+
            CRLF+'-h, --help           display help.'+
            CRLF+'-L, --longhelp       display extended help.'+
            CRLF+'-M, --merge          merge existing lossy.wav and lwcdf.wav files.'+
            CRLF+'-o, --outdir <t>     destination directory for the output file(s).'+
            CRLF+'-v, --version        display the lossyWAV version number.'+
            CRLF+'-w, --writetolog     create (or add to) lossyWAV.log in the output directory.'+CRLF;
    lossyWAV_advanced_help =
            CRLF+'Advanced Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-                    if filename="-" then WAV input is taken from STDIN.'+
            CRLF+'-t, --altpreset      enable alternative preset model which changes default'+
            CRLF+'                     behaviour regarding limit and shaping (if selected).'+
            CRLF+'-a, --analyses       set number of FFT analysis lengths, (3<=n<=5), default=2.'+
            CRLF+'    --blockdist      show distribution of lowest / highest significant bit of'+
            CRLF+'                     input codec-blocks and bit-removed codec-blocks.'+
            CRLF+'-c, --check          check if WAV file has already been processed; default=off.'+
            CRLF+'                     errorlevel=16 if already processed, 0 if not.'+
            CRLF+'-i, --impulse        force use of additional shorter FFT analysis; default=off,'+
            CRLF+'                     automatic for -q 3 and above.'+
            CRLF+'-l, --limit <n>      set upper frequency limit to be used in analyses to n Hz;'+
            CRLF+'                     (10000<=n<=20000), default=16000.'+
            CRLF+'    --linkchannels   Revert to original single bits-to-remove value for all'+
            CRLF+'                     channels rather than channel dependent bits-to-remove.'+
            CRLF+'    --maxclips <n>   set max. number of acceptable clips per channel per codec' +
            CRLF+'                     block. (0<=n<=16) Default = (3,3,3,3,2,1,0,0,0,0,0).'+
            CRLF+'-m, --midside        analyse 2 channel audio for mid/side content.'+
            CRLF+'-q, --quality <n>    quality preset (0.000<=n<=10.000); (10=highest, 0=lowest;'+
            CRLF+'                     default = --standard = 5; --insane = 10; --extreme = 7.5;'+
            CRLF+'                     --portable = 2.5; --zero = 0)'+
            CRLF+'    --sampledist     show distribution of lowest / highest significant bit of'+
            CRLF+'                     input samples and bit-removed samples.'+
            CRLF+'    --scale <n>      scaling factor from WaveGain, etc; (0.0<n<=8.0),default=1.'+
            CRLF+'-s, --shaping [n]    enable fixed noise shaping, automatic if no value input.'+
            CRLF+'                     (0.00<=n<=1.00); automatic = q/10; 0.00 = off, 1.00 = 100%'+
            CRLF+'                     effectiveness, 0.50 = 50%, etc.'+
            CRLF+'    --stdout         write processed WAV output to STDOUT.'+
            CRLF+'    --stdinname <t>  pseudo filename to use when input from STDIN.'+
            CRLF+'-U, --underlap <n>   enable underlap mode to increase number of FFT analyses'+
            CRLF+'                     performed at each FFT length, (2<=n<=8); default=2.'+
            CRLF+'-X, --sortspread     enable sort based spreading; default=off.'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'System Options:'+
            CRLF+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            CRLF+'-B, --below          set process priority to below normal.'+
            CRLF+'-d, --detail         enable detailed bits-to-remove information output mode'+
            CRLF+'    --low            set process priority to low.'+
            CRLF+'-n, --nowarnings     suppress lossyWAV warnings.'+
            CRLF+'-Q, --quiet          significantly reduce screen output.'+
            CRLF+'    --silent         no screen output.'+
            CRLF;//2345678901234567890123456789012345678901234567890123456789012345678901234567890
    lossyWAV_special_thanks =
            CRLF+'Special thanks go to:'+CRLF+
            CRLF+'David Robinson       for the publication of his lossyFLAC method, guidance, and'+
            CRLF+'                     the motivation to implement his method as lossyWAV.'+CRLF+
            CRLF+'Horst Albrecht       for ABX testing, valuable support in tuning the internal'+
            CRLF+'                     presets, constructive criticism and all the feedback.'+CRLF+
            CRLF+'Sebastian Gesemann   for the noise shaping coefficients and help in using them'+
            CRLF+'                     in the lossyWAV noise shaping implementation.'+CRLF+
            CRLF+'Matteo Frigo and     for the excellent libfftw3-3.dll contained in the FFTW'+
            CRLF+'Steven G Johnson     distribution (v3.2.1 or v3.2.2).'+CRLF+
            CRLF+'Mark G Beckett       for the Delphi unit that provides an interface to the'+
            CRLF+'(Univ. of Edinburgh) relevant fftw routines in libfftw3-3.dll.'+CRLF+
            CRLF+'Don Cross            for the Complex-FFT algorithm originally used.'+
            CRLF;//345678901234567890123456789012345678901234567890123456789012345678901234567890

    num_quality_synonyms = 4;
var
    quality_synonyms_vals     : array[0..num_quality_synonyms] of double     = ( 0.0, 2.5, 5.0, 7.5,10.0);
    quality_synonyms_short    : array[0..num_quality_synonyms] of string[3]  = ('-Z','-P','-S','-E','-I');
    quality_synonyms_long     : array[0..num_quality_synonyms] of string[11] = ('--zero','--portable','--standard','--extreme','--insane');

    this_quality_synonym_long : string;
    current_parameter         : string;
    parmError                 : string;
    logfilename_int           : String;

    wavName                   : String;
    stdinname                 : String;
    lossyName                 : String;
    lwcdfName                 : String;

    WavInpDir                 : string;
    WavOutDir                 : string;

    ThisParameterNumber       : Integer; // internal param counter for nextParamStr
    max_open_retries          : integer = 0;

    Unit_Data_Packer2         : array[0..3] of integer;

  function parmchar : char;
  begin
    parmchar:='@';
    if (length(current_parameter)=2) and (current_parameter[1]='-') then
      parmchar:=current_parameter[2];
  end;

  procedure lossyWAVerror(lwe_string : string; lwe_value : integer);
  begin
    if (not silent) then
      write(ErrOutput,'%lossyWAV Error% : ',lwe_string,CRLF);
    halt(lwe_value);
  end;

  procedure lossyWAVwarning(lww_string : string);
  begin
    if (warnings) and (not silent) then
      write(ErrOutput,'%lossyWAV Warning% : ',lww_string,CRLF);
  end;

  procedure parmsError;
  begin
    if (paramcount>0) and (Parameters.help_int=0) then
      write(ErrOutput,parmerror,CRLF)
    else
    begin
      if Parameters.help_int>1 then
        write(ErrOutput,lossyWAV_process_description+CRLF);
      write(ErrOutput,lossyWAV_standard_help);
      if Parameters.help_int>1 then
        write(ErrOutput,lossyWAV_advanced_help);
      write(ErrOutput,lossyWAV_Special_Thanks);
    end;
  end;

  function GetNextParamStr: longbool;
  begin
    GetNextParamStr:=false;
    if ThisParameterNumber<paramCount then
    begin
      ThisParameterNumber:=ThisParameterNumber + 1;
      current_parameter:=ParamStr(ThisParameterNumber);
      GetNextParamStr:=true;
    end;
  end;

  function NextParameterIsParameterOrEnd : Boolean;
  begin
    NextParameterIsParameterOrEnd:=(ThisParameterNumber=paramcount) or (copy(paramstr(ThisParameterNumber+1),1,1)='-')
  end;

  procedure parmerror_multiple_selection;
  begin
    lossyWAVerror('Multiple '+parmerror+' switches given.',$31);
  end;

  procedure parmerror_mutually_incompatible_selection;
  begin
    lossyWAVerror('Switches '+parmerror+' are incompatible.',$31);
  end;

  procedure parmerror_no_value_given;
  begin
    lossyWAVerror('No '+parmerror+' value given.',$31);
  end;

  procedure parmerror_val_error;
  begin
    lossyWAVerror('Error evaluating '+parmerror+' value given.',$31);
  end;

  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Integer); Overload;
  begin
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+inttostr(cpv_low)+'<=n<='+inttostr(cpv_high),$31);
  end;

  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Double;digits : Integer); Overload;
  begin
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+floattostrf(cpv_low,ffFixed,99,digits)+'<=n<='+floattostrf(cpv_high,ffFixed,99,digits),$31);
  end;

  function check_parameter : boolean;
  var
    qs_i, valError : Integer;
  begin
    check_parameter:=false;

    if (parmchar='v') or (current_parameter='--version') then
    begin
      write(ErrOutput,version_string,CRLF);
      halt(0);
    end;

    if (parmchar='h') or (current_parameter='--help') then
    begin
      write(ErrOutput,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      Parameters.help_int:=1;
      parmserror;
      halt(0);
    end;

    if (parmchar='L') or (current_parameter='--longhelp') then
    begin
      write(ErrOutput,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      Parameters.help_int:=2;
      parmserror;
      halt(0);
    end;

    if (parmchar='q') or (current_parameter='--quality') then
    begin
      parmerror:='quality preset';
      if Parameters.quality_int<>-99 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.quality_Int, valError);
      if valError <> 0 then
        parmerror_val_error;
      Parameters.quality_int:=nround(Parameters.quality_int,4);
      check_permitted_values(Parameters.quality_int,-4,10,4);
      if Parameters.quality_int=int(Parameters.quality_int) then
        parameter_string:='-q '+inttostr(floor(Parameters.quality_int))
      else
        parameter_string:='-q '+floattostrf(Parameters.quality_int,ffFixed,99,4);
      check_parameter:=true;
      exit;
    end;

    for qs_i:=0 to num_quality_synonyms do
    begin
      this_quality_synonym_long:=quality_synonyms_long[qs_i];
      if (current_parameter=quality_synonyms_short[qs_i]) or (current_parameter=this_quality_synonym_long) then
      begin
        parmerror:='quality preset';
        if Parameters.quality_int<>-99 then
          parmerror_multiple_selection;
        Parameters.quality_int:=quality_synonyms_vals[qs_i];
        parameter_string:=this_quality_synonym_long;
        check_parameter:=true;
      end;
    end;

    if (current_parameter='--maxclips') then
    begin
      parmerror:='maximum clips';
      if Parameters.maxclips_int<>-1 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.maxclips_Int, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.maxclips_int,0,16);
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--stdout') then
    begin
      parmerror:='output to STDOUT';
      if Parameters.STDOUTPUT_int=true then
        parmerror_multiple_selection;
      Parameters.STDOUTPUT_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='B') or (current_parameter='--below') then
    begin
      parmerror:='process priority';
      if (Parameters.priority_int>0) then
        parmerror_multiple_selection;
      setpriority($00004000);
      Parameters.priority_int:=1;
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--low') then
    begin
      parmerror:='process priority';
      if (Parameters.priority_int>0) then
        parmerror_multiple_selection;
      Parameters.priority_int:=2;
      setpriority($00000040);
      check_parameter:=true;
      exit;
    end;

    if (parmchar='d') or (current_parameter='--detail') then
    begin
      parmerror:='detail';
      if (Parameters.detail_int) then
        parmerror_multiple_selection;
      Parameters.detail_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--blockdist') then
    begin
      parmerror:='block lsb distribution';
      if (Parameters.blockdist_int) then
        parmerror_multiple_selection;
      Parameters.blockdist_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--sampledist') then
    begin
      parmerror:='sample lsb distribution';
      if (Parameters.sampledist_int) then
        parmerror_multiple_selection;
      Parameters.sampledist_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='Q') or (current_parameter='--quiet') then
    begin
      parmerror:='reduced output';
      if not Parameters.verbosity_int then
        parmerror_multiple_selection;
      Parameters.verbosity_int:=false;
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--silent') then
    begin
      parmerror:='silent running';
      if Parameters.silent_int=true then
        parmerror_multiple_selection;
      Parameters.silent_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='n') or (current_parameter='--nowarnings') then
    begin
      parmerror:='no warnings';
      if not Parameters.warnings_int then
        parmerror_multiple_selection;
      Parameters.warnings_int:=false;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='c') or (current_parameter='--check') then
    begin
      parmerror:='check';
      if Parameters.checking_int then
        parmerror_multiple_selection;
      Parameters.checking_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='M') or (current_parameter='--merge') then
    begin
      parmerror:='merge';
      if Parameters.merging_int then
        parmerror_multiple_selection;
      Parameters.merging_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='w') or (current_parameter='--writetolog') then
    begin
      parmerror:='write log output';
      if Parameters.writetolog_int then
        parmerror_multiple_selection;
      Parameters.writetolog_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--stdinname') then
    begin
      parmerror:='pseudo filename';
      if stdinname<>'' then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      stdinname:=current_parameter;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='f') or (current_parameter='--force') then
    begin
      parmerror:='forced over-write';
      if Parameters.forcing_int then
        parmerror_multiple_selection;
      Parameters.forcing_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='o') or (current_parameter='--outdir') then
    begin
      parmerror:='destination folder';
      if (WAVOutDir <> '') then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      WAVOutDir:=current_parameter;
      if not((WAVOutDir[length(WAVOutDir)]='\') or (WAVOutDir[length(WAVOutDir)]=':')) then
        WAVOutdir:=WAVOutdir+'\';
      check_parameter:=true;
      exit;
    end;

    if (parmchar='C') or (current_parameter='--correction') then
    begin
      parmerror:='correction file';
      if Parameters.correct_int then
        parmerror_multiple_selection;
      Parameters.correct_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='t') or (current_parameter='--altpreset') then
    begin
      parmerror:='altpreset';
      if Parameters.altpreset_int then
        parmerror_multiple_selection;

      Parameters.altpreset_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='s') or (current_parameter='--shaping') then
    begin
      parmerror:='noise shaping';
      if Parameters.shaping_int<>-1 then
        parmerror_multiple_selection;
      if NextParameterIsParameterOrEnd then
        Parameters.shaping_int:=-2
      else
      begin
        if not GetNextParamStr then
          parmerror_no_value_given;

        Val(current_parameter, Parameters.shaping_int, valError);
        if valError <> 0 then
          parmerror_val_error;

        Parameters.shaping_int:=nround(Parameters.shaping_int,4);
        check_permitted_values(Parameters.shaping_int,0.0,1.0,4);
      end;
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--scale') then
    begin
      parmerror:='scaling factor';
      if Parameters.scaling_int<>-1 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.scaling_int, valError);
      if valError <> 0 then
        parmerror_val_error;

      Parameters.scaling_int:=nround(Parameters.scaling_int,5);
      check_permitted_values(Parameters.scaling_int,PowersOf.ten[-5],8.0,5);
      check_parameter:=true;
      exit;
    end;

    if (parmchar='l') or (current_parameter='--limit') then
    begin
      parmerror:='limit frequency';
      if Parameters.limit_int<>-1 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.limit_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.limit_int,10000,20000);
      check_parameter:=true;
      exit;
    end;

    if (parmchar='a') or (current_parameter='--analyses') then
    begin
      parmerror:='number of FFT analyses';
      if Parameters.analyses_int<>2 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.analyses_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.analyses_int,3,5);
      check_parameter:=true;
      exit;
    end;

    if (parmchar='i') or (current_parameter='--impulse') then
    begin
      parmerror:='impulse';
      if Parameters.impulse_int then
        parmerror_multiple_selection;
      Parameters.impulse_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='m') or (current_parameter='--midside') then
    begin
      parmerror:='midside';
      if Parameters.midside_int then
        parmerror_multiple_selection;
      Parameters.midside_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (current_parameter='--linkchannels') then
    begin
      parmerror:='link channels';
      if Parameters.linkchannels_int then
        parmerror_multiple_selection;
      Parameters.linkchannels_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='X') or (current_parameter='--sortspread') then
    begin
      parmerror:='sortspread';
      if Parameters.sortspread_int then
        parmerror_multiple_selection;
      Parameters.sortspread_int:=true;
      check_parameter:=true;
      exit;
    end;

    if (parmchar='U') or (current_parameter='--underlap') then
    begin
      parmerror:='underlap';
      if Parameters.underlap_int<>0 then
        parmerror_multiple_selection;
      if not GetNextParamStr then
        parmerror_no_value_given;
      Val(current_parameter, Parameters.underlap_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(Parameters.underlap_int,2,8);
      check_parameter:=true;
      exit;
    end;
  end;

  function getParms: longbool;
  var
    colonpos : Integer;
    slashpos : Integer;
  begin
    getParms        :=false;

    ThisParameterNumber:=0;

    WAVInpDir       :='';
    WAVOutDir       :='';
    wavName         :='';
    stdinname       :='';

    Parameters.priority_int    :=0;
    Parameters.quality_int     :=-99;
    Parameters.scaling_int     :=-1;
    Parameters.shaping_int     :=-1;

    Parameters.analyses_int    :=2;
    Parameters.maxclips_int    :=-1;
    Parameters.help_int        :=0;

    Parameters.limit_int     :=-1;
    Parameters.underlap_int    :=0;

    Parameters.sortspread_int  :=false;
    Parameters.blockdist_int   :=false;
    Parameters.checking_int    :=false;
    Parameters.correct_int     :=false;

    Parameters.detail_int      :=false;
    Parameters.forcing_int     :=false;
    Parameters.impulse_int     :=false;
    Parameters.linkchannels_int:=false;

    Parameters.merging_int     :=false;
    Parameters.midside_int     :=false;
    Parameters.sampledist_int  :=false;
    Parameters.silent_int      :=false;

    Parameters.STDINPUT_int    :=false;
    Parameters.STDOUTPUT_int   :=false;
    Parameters.warnings_int    :=true;
    Parameters.writetolog_int  :=false;

    Parameters.verbosity_int   :=true;
    Parameters.altpreset_int   :=false;

    if not GetNextParamStr then
      exit;

    if (current_parameter='-') then
      Parameters.STDINPUT_int:=true
    else
      if (current_parameter[1]='-') then
      begin
        if (parmchar='v') or (current_parameter='--version') then
        begin
          write(ErrOutput,version_string,CRLF);
          halt(0);
        end
        else
        begin
          write(ErrOutput,lossyWAVhead1,CRLF);
          write(ErrOutput,lossyWAV_GPL,CRLF);
          if (parmchar='h') or (current_parameter='--help') or (parmchar='L') or (current_parameter='--longhelp') then
          begin
            Parameters.help_int:=1+ord((current_parameter='--longhelp') or (parmchar='L'));
            parmserror;
            halt(0);
          end
          else
            lossyWAVerror('No input file given.',1);
        end;
      end
      else
        wavName:=current_parameter;

    if (wavname<>'') then
    begin
      if (wavName[1]='"') then
        wavName:=copy(wavName,2,length(wavName)-1);
      if (wavName[length(wavName)]='"') then
        wavName:=copy(wavName,1,length(wavName)-1);

      colonpos:=pos(':',wavName);
      if colonpos>0 then
      begin
        wavInpDir:=copy(wavName,1,colonpos);
        wavName:=copy(wavName,colonpos+1,length(wavName)-colonpos);
      end;

      slashpos:=pos('\',wavName);
      while slashpos>0 do
      begin
        wavInpDir:=wavInpDir+copy(wavName,1,slashpos);
        wavName:=copy(wavName,slashpos+1,length(wavName)-slashpos);
        slashpos:=pos('\',wavName);
      end;
    end;

    while GetNextParamStr do
    begin
      if not check_parameter then
      begin
        parmError:='%lossyWAV Error% : Incorrect option: "' + current_parameter + '"'+CRLF;
        exit;
      end;
    end;
    getParms:=true;
  end;

  //============================================================================

  function DirInp : String;
  begin
    DirInp:=wavInpDir;
  end;

  function DirOut : String;
  begin
    DirOut:=wavOutDir;
  end;

  Function wavFilePrintName : String;
  Begin
    if STDINPUT then
    begin
      if stdinname='' then
        wavFilePrintName:='from STDIN'
      else
        wavfileprintname:=stdinname;
    end
    else
      wavFilePrintName:=wavname;
  End;

  Function lossyFileName : String;
  Begin
    lossyFileName:=lossyName;
  End;

  Function lwcdfFileName : String;
  Begin
    lwcdfFileName:=lwcdfName;
  End;

  function wavInp: string;
  begin
    wavInp:=wavInpDir + wavName;
  end;

  Function WAVOut: String;
  Begin
    WAVOut:=WAVOutDir + WAVName;
  End;

  function lossyInp: String;
  begin
    lossyInp:=WAVInpDir + lossyFileName;
  end;

  function lossyOut: string;
  begin
    lossyOut:=WAVOutDir + lossyFileName;
  end;

  function lwcdfInp: String;
  begin
    lwcdfInp:=WAVInpDir + lwcdfFileName;
  end;

  function lwcdfOut: string;
  begin
    lwcdfOut:=WAVOutDir + lwcdfFileName;
  end;

  function outputname : String;
  begin
    outputname:=stdinname;
  end;

  function logfilename : String;
  begin
    logfilename:=logfilename_int;
  end;

  function quality: Double;
  begin
    quality:=Parameters.quality_int;
  end;

  function shaping : Double;
  begin
    shaping:=Parameters.shaping_int;
  end;

  function limit : Integer;
  begin
    limit:=Parameters.limit_int;
  end;

  function scaling : Double;
  begin
    scaling:=Parameters.scaling_int;
  end;

  function maxclips : integer;
  begin
    maxclips:=Parameters.maxclips_int;
  end;

  function priority : Integer;
  begin
    priority:=Parameters.priority_int;
  end;

  function analyses : Integer;
  begin
    analyses:=Parameters.analyses_int;
  end;

  function forcing : longbool;
  begin
    forcing:=Parameters.forcing_int;
  end;

  function correct : longbool;
  begin
    correct:=Parameters.correct_int;
  end;

  function checking : longbool;
  begin
    checking:=Parameters.checking_int;
  end;

  function merging : longbool;
  begin
    merging:=Parameters.merging_int;
  end;

  function silent : longbool;
  begin
    silent:=Parameters.silent_int;
  end;

  function verbosity : longbool;
  begin
    verbosity:=Parameters.verbosity_int;
  end;

  function altpreset : longbool;
  begin
    altpreset:=Parameters.altpreset_int;
  end;

  function detailedoutput : longbool;
  begin
    detailedoutput:=Parameters.detail_int;
  end;

  function warnings : longbool;
  begin
    warnings:=Parameters.warnings_int;
  end;

  function writetolog : longbool;
  Begin
    writetolog:=Parameters.writetolog_int;
  End;

  function STDINPUT : longbool;
  begin
    STDINPUT:=Parameters.STDINPUT_int;
  end;

  function STDOUTPUT : longbool;
  begin
    STDOUTPUT:=Parameters.STDOUTPUT_int;
  end;

  function blockdist : longbool;
  begin
    blockdist:=Parameters.blockdist_int;
  end;

  function sampledist : longbool;
  begin
    sampledist:=Parameters.sampledist_int;
  end;

  function linkchannels : longbool;
  begin
    linkchannels:=Parameters.linkchannels_int;
  end;

  function impulse : longbool;
  begin
    impulse:=Parameters.impulse_int;
  end;

  function midside : longbool;
  begin
    midside:=Parameters.midside_int;
  end;

  function underlap : integer;
  begin
    underlap:=Parameters.underlap_int;
  end;

  function sortspread : longbool;
  begin
    sortspread:=Parameters.sortspread_int;
  end;

  {$I-}
  procedure open_log_file;
  var
    open_retries : Integer;
    lastioresult : Integer;
  begin
    Open_Retries:=0;
    if not Parameters.logfileopened_int then
    begin
      filemode:=1;                                                  //write only
      assign(LogOutput,logfilename_int);
      if fileexists(logfilename_int) then
      begin
        if filegetattr(logfilename_int) and 1=1 then
        Begin
          lossyWAVwarning('lossyWAV.log file is read-only.');
          if(filesetattr(logfilename_int,filegetattr(logfilename_int) and $FE)<>0) then
            lossyWAVerror('Cannot gain write access to '+logfilename_int,$31);
        End;
        Repeat
          {$I-}
          Append(LogOutput);
          lastIOResult:=ioresult;
          {$I+}
          if lastioresult<>0 then
          begin
            inc(open_retries);
            timer.DelayTime:=PowersOf.two[-5];
            ndelay;
          end;
        Until (lastioresult=0) or (open_retries=128);
        if lastioresult<>0 then
          lossywaverror('Opening lossyWAV.log',$31);
        max_open_retries:=max(open_retries,max_open_retries);
      end
      else
      begin
        {$I-}
        Rewrite(LogOutput);
        lastioresult:=ioresult;
        {$I+}
        if lastioresult<>0 then
          lossywaverror('Opening lossyWAV.log',$31);
      end;
    end;
    Parameters.logfileopened_int:=true;
  end;

  procedure close_log_file;
  begin
    if Parameters.logfileopened_int then
    begin
      flush(LogOutput);
      close(LogOutput);
    end;
    Parameters.logfileopened_int:=false;
  end;
  {$I+}

  //============================================================================
  procedure nParameter_Init;
  //============================================================================
  begin
    Parameters.parameters_checked_int:=false;
    Parameters.logfileopened_int:=false;

    {$I-}
    if (not getparms) then
    begin
      write(ErrOutput,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      if parmerror<>'' then
        write(ErrOutput,parmerror);
      write(ErrOutput,'Syntax: lossyWAV <wavfilename>'+CRLF);
      write(ErrOutput,CRLF+'Type ''lossyWAV --help'' for more detail.'+CRLF);
      halt(1);
    end;

    if Parameters.quality_int=-99 then
    begin
      Parameters.quality_int:=5;
      parameter_string:='--standard';
    end;

    if (not altpreset()) and (parameters.quality_int<0) then
    begin
      Parameters.quality_int:=0;
      parameter_string:='--zero';
    end;

    if (Parameters.help_int>=1) then
    begin
      write(ErrOutput,lossyWAVhead1,CRLF);
      write(ErrOutput,lossyWAV_GPL,CRLF);
      parmserror;
      halt(1);
    end;

    if (verbosity) and (not silent) then
      write(ErrOutput,lossyWAVhead1,lossywavhead2);

    if (not STDINPUT) and (length(wavName)=0) then
      lossyWAVerror('Name of input or output wav file missing.',$31);

    if (STDINPUT) then
    begin
      wavname:='';
      lossyName:='lossywav.lossy.wav';
      lwcdfname:='lossywav.lwcdf.wav';
      if (Parameters.merging_int) then
        lossyWAVerror('Merge parameter is incompatible'+CRLF+'                   with STDIN file input mode.',$31);
    end
    else
    begin
      if not fileexists(wavInp) then
      begin
        if not fileexists(wavInp+'.wav') then
          lossyWAVerror('Input file: '+wavName+' does not exist.',$31)
        else
          wavName:=wavName+'.wav';
      end;
    end;

    if (Parameters.STDOUTPUT_int=true) then
    begin
      lossyname:='';
      lwcdfname:='';
      wavoutdir:='';
      if (Parameters.correct_int) then
        lossyWAVerror('Correct parameter is incompatible'+CRLF+'                   with STDOUT file output mode.',$31);
    end
    else
    begin
      if Parameters.stdinput_int then
      begin
        if stdinname<>'' then
        begin
          if (length(stdinname)<5) then
            if copy(stdinname,length(stdinname),1)='.' then
              stdinname:=stdinname+'wav'
            else
              stdinname:=stdinname+'.wav';

          lossyName:=copy(stdinname, 1, length(stdinname) - 3) + 'lossy.' + copy(stdinname, length(stdinname) - 2, 3);
          lwcdfName:=copy(stdinname, 1, length(stdinname) - 3) + 'lwcdf.' + copy(stdinname, length(stdinname) - 2, 3);
        end
        else
        begin
          lossyname:='stdin.lossy.wav';
          lwcdfname:='stdin.lwcdf.wav';
        end;
      end
      else
      begin
        lossyName:=copy(wavName, 1, length(wavName) - 3) + 'lossy.' + copy(wavName, length(wavName) - 2, 3);
        lwcdfName:=copy(wavName, 1, length(wavName) - 3) + 'lwcdf.' + copy(wavName, length(wavName) - 2, 3);
      end;

      if (WAVOutdir<>'') and (directoryexists(WAVOutdir)=false) then
        lossyWAVerror('Directory : '+WAVOutdir+' cannot be accessed.',$31);
    end;

    if not Parameters.merging_int then
    begin
      if (Parameters.STDOUTPUT_int=false) and (fileexists(lossyOut)=true) then
      begin
        if Parameters.forcing_int=true then
        Begin
          if (filegetattr(lossyOut) and 1)=1 then
          Begin
            lossyWAVwarning('Output lossy file is read-only.');
            if(filesetattr(lossyOut,filegetattr(lossyOut) and $FE)<>0) then
              lossyWAVerror('Cannot gain write access to '+lossyOut,$31);
          End;
        End
        else
          lossyWAVerror('Output file: '+lossyOut+' exists.',$31);
      end;

      if (correct=true) and (fileexists(lwcdfOut)=true) then
      begin
        if (forcing=true) then
        Begin
          if filegetattr(lwcdfOut) and 1=1 then
          Begin
            lossyWAVwarning('Output correction file is read-only.');
            if(filesetattr(lwcdfOut,filegetattr(lwcdfOut) and $FE)<>0) then
              lossyWAVerror('Cannot gain write access to '+lwcdfOut,$31);
          End;
        End
        else
          lossyWAVerror('Output file: '+lwcdfOut+' exists.',$31);
      end;
    end
    else
    begin
      lossyName:=wavName;
      if (copy(lossyName,length(lossyName)-9,10)='.lossy.wav') then
      begin
        lwcdfName:=copy(wavName,1,length(wavName)-10)+'.lwcdf.wav';
        wavName:=copy(wavName,1,length(wavName)-10)+'.wav';
      end;

      if fileexists(WAVOut) and (not forcing) then
        lossyWAVerror('Output merged file already exists',$31);

      if not fileexists(lossyInp) then
        lossyWAVerror('Input lossy file not found.',$31);

      if not fileexists(lwcdfInp) then
        lossyWAVerror('Input correction file not found.',$31);
    end;

    logfilename_int:=DirOut+'lossyWAV.log';

    {$I+}
  //============================================================================
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('nParameter',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}

  nParameter_Init;
//==============================================================================
end.
//==============================================================================

