//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nProcess;
//==============================================================================
interface
//==============================================================================
var
  Unit_Data_Packer1 : Array[0..15] of Integer;
//==============================================================================
Procedure Process_This_Codec_Block;
//==============================================================================
implementation
//==============================================================================
{$DEFINE SAMPLEDIST_ON}
//==============================================================================
uses
  fftw_interface,
  {$IFDEF USEASM}
  nFFT_asm,
  {$ELSE}
  nFFT_pas,
  {$ENDIF}
  nFillFFT,
  nMaths,
  nOutput,
  nParameter,
  nRemoveBits,
  nSpreading,
  nSystem,
  nWav,
  math,
  SysUtils;
//==============================================================================
var
  Unit_Data_Packer2 : array[0..14] of integer;

  //============================================================================
  procedure process_this_codec_block;
  //============================================================================
  var
    this_channel                          : Integer;
    this_analysis_number                  : Integer;
    this_analysis_block_number            : Integer;
    this_analysis_block_start             : Integer;

    codec_block_dependent_bits_to_remove  : Integer;
    min_neg_codec_block_start             : Integer;
    max_pos_codec_block_start             : Integer;
    local_channels                        : Integer;

    Spreading_Old                         : Double;
    Spreading_New                         : Double;
    Spreading_result                      : Double;

    bits_removed_this_codec_block         : Double;
    this_fft_underlap_length              : Double;
    check_channel_source_result           : Double;
    check_channel_correction_result       : Double;

    channel_data                          : Array[0..MAX_CHANNELS-1] of channel_data_rec;
  begin
    codec_block_dependent_bits_to_remove:=bitspersample;
    bits_removed_this_codec_block:=0;

    if midside() and (Global.channels=2) then
      local_channels:=4
    else
      local_channels:=Global.channels;

    //==========================================================================
    for this_channel:=0 to local_channels-1 do
    //==========================================================================
    begin
      Global.channel:=this_channel;

      channel_data[Global.channel].maximum_bits_to_remove:=max(0,floor(this_codec_block_channel_log2rms[Global.channel]-settings.dynamic_minimum_bits_to_keep));
      channel_data[Global.channel].maximum_bits_to_remove:=min(channel_data[Global.channel].maximum_bits_to_remove,settings.static_maximum_bits_to_remove);

      Results.This_FFT.btr:=channel_data[Global.channel].maximum_bits_to_remove;

      channel_data[Global.channel].min_fft_result.btr:=channel_data[Global.channel].maximum_bits_to_remove;

      for this_analysis_number:=1 to PRECALC_ANALYSES do
      begin
        Global.analysis_number:=this_analysis_number;
        Set_Global_FFT_Variables(global.codec_block_bits-5+Global.Analysis_Number);

        if ((settings.fft_analysis_switches and Global.fft_length)>0) and (Global.fft_length<=Global.codec_block_size*2)then
        begin
          if not fftw_initialised then
            if SetFFTLength(PowersOf.twoint[Global.fft_bit_length-1])=0 then
              lossyWAVerror('Fatal FFT error!',$88);

          min_neg_codec_block_start:=-(last_codec_block_size+prev_codec_block_size);
          max_pos_codec_block_start:=this_codec_block_size+next_codec_block_size-Global.fft_length;

          this_analysis_block_start:=process.actual_analysis_blocks_start[Global.fft_bit_length];
          this_fft_underlap_length:=process.fft_underlap_length[Global.fft_bit_length];

          for this_analysis_block_number:=0 to process.analysis_blocks[Global.fft_bit_length] do
          begin
            actual_analysis_block_start:=floor(this_analysis_block_start+this_analysis_block_number*this_fft_underlap_length);
            actual_analysis_block_start:=min(max(actual_analysis_block_start,min_neg_codec_block_start),max_pos_codec_block_start);

            if (this_analysis_block_number=0) then
            begin
              if (actual_analysis_block_start=Results.saved_fft_results[Global.channel,Global.analysis_number].start) then
              begin
                Results.This_FFT:=Results.saved_fft_results[Global.channel,Global.analysis_number];
                if (Results.This_FFT.btr<channel_data[Global.channel].min_fft_result.btr) or (channel_data[Global.channel].min_fft_result.btr=-1) then
                  channel_data[Global.channel].min_fft_result:=Results.This_FFT;
              end;
            end
            else
            begin
              if Fill_FFT_Input()=0 then
                Results.This_FFT.btr:=0
              else
              begin
                if FFTW_Initialised then
                  FFTW_EXECUTE(fftw_plans[Global.fft_bit_length])
                else
                  FFT_DIT_Real();

                Spreading_Function();

                Spreading_old:=min(Spreading.old_minimum+settings.noise_threshold_shift_minimum,Spreading.old_average+settings.noise_threshold_shift_average);

                Spreading_new:=min(Spreading.new_minimum+settings.noise_threshold_shift_minimum,Spreading.new_average+settings.noise_threshold_shift_average);

                Spreading_result:=min(Spreading_Old,Spreading_New);

                Results.This_FFT.spreading:=max(0,Spreading_result-window_function_RMS);

                Spreading_Result:=Min(THRESHOLD_INDEX_SPREAD_MUL-1,Results.This_FFT.spreading)*THRESHOLD_INDEX_SPREAD;

                Results.This_FFT.btr:=threshold_index[Global.fft_bit_length,floor(Spreading_result)];
              end;

              if (Results.This_FFT.btr<channel_data[Global.channel].min_fft_result.btr) or (channel_data[Global.channel].min_fft_result.btr=-1) then
                channel_data[Global.channel].min_fft_result:=Results.This_FFT;
            end;
          end;
          Results.saved_fft_results[Global.channel,Global.analysis_number].start:=actual_analysis_block_start-Global.codec_block_size;
          Results.saved_fft_results[Global.channel,Global.analysis_number].btr:=Results.This_FFT.btr;
          Results.saved_fft_results[Global.channel,Global.analysis_number].spreading:=Results.This_FFT.spreading;
        end;
      end;

      if channel_data[Global.channel].min_fft_result.btr<0 then
        channel_data[Global.channel].min_fft_result.btr:=0;

      channel_data[Global.channel].bits_to_remove:=channel_data[Global.channel].min_fft_result.btr;

      if Global.channel<Global.channels then
      begin
        remove_bits_data.bits_to_remove:=channel_data[Global.channel].bits_to_remove;
        remove_bits();
        channel_data[Global.channel].bits_to_remove:=remove_bits_data.bits_removed;
        channel_data[Global.channel].clips:=remove_bits_data.clips;
        codec_block_dependent_bits_to_remove:=min(codec_block_dependent_bits_to_remove,channel_data[Global.channel].bits_to_remove);
      end;
    //==========================================================================
    end;
    //==========================================================================

    //==========================================================================
    for this_channel:=0 to Global.channels-1 do
    //==========================================================================
    begin
      Global.channel:=this_channel;

      if linkchannels or midside then
      begin
        remove_bits_data.bits_to_remove:=codec_block_dependent_bits_to_remove;
        if channel_data[Global.channel].bits_to_remove<>remove_bits_data.bits_to_remove then
        begin
          if Global.channel<Global.channels then
            remove_bits();
          channel_data[Global.channel].bits_to_remove:=remove_bits_data.bits_removed;
          channel_data[Global.channel].clips:=remove_bits_data.clips;
        end;
      end;

      channel_data[Global.channel].bits_lost:=channel_data[Global.channel].min_fft_result.btr-channel_data[Global.channel].bits_to_remove;

      Global.bits_removed:=Global.bits_removed+channel_data[Global.channel].bits_to_remove;
      Global.bits_lost:=Global.bits_lost+channel_data[Global.channel].bits_lost;
      Global.clipped_samples:=Global.clipped_samples+channel_data[Global.channel].clips;
      
      if detailedoutput then
        bit_removal_history^[(Global.blocks_processed-1)*Global.channels+Global.channel]:=channel_data[Global.channel].bits_to_remove;

      bits_removed_this_codec_block:=bits_removed_this_codec_block+OneOver[Global.channels]*channel_data[Global.channel].bits_to_remove;
    //==========================================================================
    end;
    //==========================================================================

    //==========================================================================
    // copy FIFO saved state from temp storage to saved storage,
    // i.e. no more bit-removal for this codec-block.
    //==========================================================================
    fifo_save_state();
    //==========================================================================

    Global.samples_processed:=(Global.blocks_processed-1)*Global.codec_block_size+this_codec_block_size;

    Process_Output;

    {$IFDEF SAMPLEDIST_ON}
    if blockdist() or sampledist() then
      lsb_analysis();
    {$ENDIF}
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('nProcess',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
