//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nRemoveBits;
//==============================================================================
interface
//==============================================================================
uses
  nSystem,
  nwav;
//==============================================================================
var
  Unit_Data_Packer1 : Array[0..15] of Integer;
//==============================================================================
const
  order_4_44100                         : tShaping = (A:(+1.05870,+0.06760,-0.60540,-0.27380); B:(-2.20610,+0.47070,+0.25340,+0.62130));
  order_4_48000                         : tShaping = (A:(+0.90300,+0.01160,-0.58530,-0.25710); B:(-2.23740,+0.73390,+0.12510,+0.60330));
//==============================================================================
var
  remove_bits_data                      : packed record
    bits_to_remove  : Integer;
    clips           : Integer;
    bits_removed    : Integer;
  end;

  //============================================================================
  // Max sample lookup table for rounding.
  //============================================================================
  max_samples                           : array[0..BITS_TO_CALCULATE-1] of Integer;

//==============================================================================
procedure fifo_save_state();
Procedure remove_bits();
Procedure Remove_Bits_Init();
//==============================================================================
implementation
//==============================================================================
uses
  nparameter,
  math;
//==============================================================================
var
  Shaping_Array                         : tShaping;
  fifo_multiplier                       : Double;
  fifo_multiplier_multiplier            : Double;
  this_min_sample                       : Integer;
  this_max_sample                       : Integer;

  fifo                                  : array[-shaping_length..MAX_BLOCK_SIZE-1] of double;
  saved_fifo_state                      : array[0..MAX_CHANNELS-1,0..shaping_length-1] of double;
  temp_fifo_state                       : array[0..MAX_CHANNELS-1,0..shaping_length-1] of double;

  shaping_factor_power                  : Array[0..shaping_length-1] of double;

  Unit_Data_Packer2                     : array[0..5] of integer;

//==============================================================================
  {$IFNDEF USEASM}
  //----------------------------------------------------------------------------
  Procedure Remove_Bits_Proc_Shaping_On;
  //----------------------------------------------------------------------------
  Var
    rb_i, rb_j : integer;
    rb_k : int64;
    output_temp : double;
    wanted_temp : double;
    this_two_power : Integer;
    this_two_power_recip : Double;
  Begin
    remove_bits_data.clips:=999;

    while (remove_bits_data.bits_removed>-1) and (remove_bits_data.clips>settings.maximum_clips_per_channel) do
    Begin
      remove_bits_data.clips:=0;
      this_max_sample:=max_samples[remove_bits_data.bits_removed];
      this_two_power:=PowersOf.twoint[remove_bits_data.bits_removed];
      this_two_power_recip:=PowersOf.two[-remove_bits_data.bits_removed];

      if remove_bits_data.bits_removed=0 then
        fifo_multiplier_multiplier:=single_half
      else
        fifo_multiplier_multiplier:=powersof.two[0];

      fifo_multiplier:=powersof.two[0];

      for rb_i:=0 to this_codec_block_size-1 do
      begin
        rb_j:=floor((rotating_blocks_ptr[2]^[Global.channel,rb_i])*settings.scaling_factor);

        wanted_temp:=rb_j
                    +fifo[rb_i-1]*shaping_array.b[0]
                    +fifo[rb_i-2]*shaping_array.b[1]
                    +fifo[rb_i-3]*shaping_array.b[2]
                    +fifo[rb_i-4]*shaping_array.b[3];

        output_temp:=round(wanted_temp*this_two_power_recip)*this_two_power;

        fifo[rb_i]:=((output_temp-wanted_temp)
                                 -fifo[rb_i-1]*shaping_array.a[0]
                                 -fifo[rb_i-2]*shaping_array.a[1]
                                 -fifo[rb_i-3]*shaping_array.a[2]
                                 -fifo[rb_i-4]*shaping_array.a[3])*fifo_multiplier;

        fifo_multiplier:=fifo_multiplier*fifo_multiplier_multiplier;

        rb_k:=floor(output_temp);

        if rb_k>this_max_sample then
        begin
          rb_k:=this_max_sample;
          inc(remove_bits_data.clips);
        end
        else
          if rb_k<this_min_sample then
          begin
            rb_k:=min_sample;
            inc(remove_bits_data.clips);
          end;

        rb_j:=rb_j-round(rb_k*settings.scaling_factor_inv);
        btrd_codec_block[Global.channel,rb_i]:=rb_k;
        corr_codec_block[Global.channel,rb_i]:=rb_j;
      end;

      remove_bits_data.bits_removed:=remove_bits_data.bits_removed-ord(remove_bits_data.clips>settings.maximum_clips_per_channel);
    end;

    if remove_bits_data.bits_removed<0 then
      remove_bits_data.bits_removed:=0;
  End;
  //----------------------------------------------------------------------------

  //----------------------------------------------------------------------------
  Procedure Remove_Bits_Proc_Shaping_Off;
  //----------------------------------------------------------------------------
  Var
    rb_i, rb_j : integer;
    rb_k : int64;
    this_two_power : Integer;
    this_two_power_recip : Double;
  Begin
    remove_bits_data.clips:=999;

    while (remove_bits_data.bits_removed>-1) and (remove_bits_data.clips>settings.maximum_clips_per_channel) do
    Begin
      remove_bits_data.clips:=0;
      this_max_sample:=max_samples[remove_bits_data.bits_removed];
      this_two_power:=PowersOf.twoint[remove_bits_data.bits_removed];
      this_two_power_recip:=PowersOf.two[-remove_bits_data.bits_removed];

      if remove_bits_data.bits_removed=0 then
        fifo_multiplier_multiplier:=powersof.two[-1]
      else
        fifo_multiplier_multiplier:=powersof.two[0];

      fifo_multiplier:=powersof.two[0];

      for rb_i:=0 to this_codec_block_size-1 do
      begin
        rb_j:=floor(rotating_blocks_ptr[2]^[Global.channel,rb_i]*settings.scaling_factor);
        rb_k:=round(rb_j*this_two_power_recip)*this_two_power;

        if rb_k>this_max_sample then
        begin
          rb_k:=this_max_sample;
          inc(remove_bits_data.clips);
        end
        else
          if rb_k<this_min_sample then
          begin
            rb_k:=min_sample;
            inc(remove_bits_data.clips);
          end;

        rb_j:=rb_j-round(rb_k*settings.scaling_factor_inv);
        btrd_codec_block[Global.channel,rb_i]:=rb_k;
        corr_codec_block[Global.channel,rb_i]:=rb_j;
      end;

      remove_bits_data.bits_removed:=remove_bits_data.bits_removed-ord(remove_bits_data.clips>settings.maximum_clips_per_channel);
    end;

    if remove_bits_data.bits_removed<0 then
      remove_bits_data.bits_removed:=0;
  End;
  //----------------------------------------------------------------------------
  {$ELSE}
  //============================================================================
  procedure remove_bits_proc_shaping_on; assembler;
  //============================================================================
  asm
                push    ebx
                push    ebp
                push    esi
                push    edi

      @@01:       mov     eax,dword ptr remove_bits_data.bits_removed
                  cmp     eax,0
                  jl      @@29

                  fld1
                  fst     fifo_multiplier
                  jg      @@03

                  fmul    single_half

      @@03:       fstp    fifo_multiplier_multiplier

                  sub     esi,esi
                  fld     dword ptr PowersOf.Two[128*SINGLE_SIZE-eax*SINGLE_SIZE]
                  neg     eax
                  fld     dword ptr PowersOf.two[128*SINGLE_SIZE+eax*SINGLE_SIZE]
                  neg     eax
                  fild    dword ptr max_samples[eax*INTEGER_SIZE]
                  fild    this_min_sample

                  mov     ebx,Global.channel
                  mov     ebx,dword ptr channel_offset[ebx*POINTER_SIZE]
                  mov     ebp,dword ptr rotating_blocks_ptr[2*POINTER_SIZE]

                  mov     edi,this_codec_block_size
                  shl     edi,integer_shift
                  add     edi,ebx
                  mov     ecx,shaping_length*DOUBLE_SIZE

      @@09:         //==========================================================
                    // Start of bit-removal
                    //==========================================================
      @@10:         fild    dword ptr [ebp+ebx]                        // top of FPU stack = current sample value
                    fmul    settings.scaling_factor

                    //==========================================================
                    // Noise-shaped bit-removal.
                    //==========================================================
                    // wanted_temp = x[k] + b[1*]*fifo[1] + b[2*]*fifo[2] + b[3*]*fifo[3] + ... + b[n*]*fifo[n]
                    //==========================================================
      @@12:         fld     qword ptr shaping_array[(shaping_length+0)*DOUBLE_SIZE] // b[1*]
                    fmul    qword ptr fifo[ecx-1*DOUBLE_SIZE]          // fifo[1]
                    fadd
                    fld     qword ptr shaping_array[(shaping_length+1)*DOUBLE_SIZE] // b[2*]
                    fmul    qword ptr fifo[ecx-2*DOUBLE_SIZE]          // fifo[2]
                    fadd
                    fld     qword ptr shaping_array[(shaping_length+2)*DOUBLE_SIZE] // b[3*]
                    fmul    qword ptr fifo[ecx-3*DOUBLE_SIZE]          // fifo[3]
                    fadd
                    fld     qword ptr shaping_array[(shaping_length+3)*DOUBLE_SIZE] // b[4*]
                    fmul    qword ptr fifo[ecx-4*DOUBLE_SIZE]          // fifo[4]
                    fadd

                    //==========================================================
                    // make a copy of the newly calculated wanted_temp value
                    //==========================================================
                    fld     st                                         // wanted_temp

                    fmul    st,st(4)                                   // divide by 2^(bits-to-remove)
                    sub     esp,4
                    fistp   dword ptr [esp]                            // y[k] = round(wanted_temp); frndint replacement
                    fild    dword ptr [esp]                            // code - thanks Agner Fog!
                    add     esp,4
                    fmul    st,st(5)                                   // multiply by 2^(bits-to-remove)

                    fxch    st(1)                                      // swap y[k] and wanted_temp
                    fsubr   st,st(1)                                   // qerror_temp = wanted_temp - y[k]

                    //==========================================================
                    // qerror_temp is on top of the FPU stack.
                    // new fifo = qerror_temp - a[1*]*fifo[1] - a[2*]*fifo[2] - a[3*]*fifo[3] - ... - a[n*]*fifo[n]
                    //==========================================================
                    fld     qword ptr shaping_array[(0+0)*DOUBLE_SIZE] // a[1*]
                    fmul    qword ptr fifo[ecx-1*DOUBLE_SIZE]          // fifo[1]
                    fsub
                    fld     qword ptr shaping_array[(0+1)*DOUBLE_SIZE] // a[2*]
                    fmul    qword ptr fifo[ecx-2*DOUBLE_SIZE]          // fifo[2]
                    fsub
                    fld     qword ptr shaping_array[(0+2)*DOUBLE_SIZE] // a[3*]
                    fmul    qword ptr fifo[ecx-3*DOUBLE_SIZE]          // fifo[3]
                    fsub
                    fld     qword ptr shaping_array[(0+3)*DOUBLE_SIZE] // a[4*]
                    fmul    qword ptr fifo[ecx-4*DOUBLE_SIZE]          // fifo[4]
                    fsub

                    fmul    fifo_multiplier
                    fstp    qword ptr fifo[ecx]                        // new fifo

                    fld     fifo_multiplier
                    fmul    fifo_multiplier_multiplier
                    fstp    fifo_multiplier
                    //==========================================================
                    // End of noise shaping code.
                    //==========================================================

      @@16:         //==========================================================
                    // Check that new value lies within acceptable limits.
                    //==========================================================
                    fcom    st(2)
                    fnstsw  ax
                    and     ah,$41
                    jnz     @@18

                    fstp    st
                    fld     st(1)
                    add     esi,1
                    jmp     @@20

      @@18:         fcom    st(1)                  // min_sample
                    fnstsw  ax
                    xor     ah,$01
                    and     ah,$41
                    jnz     @@20

                    fstp    st
                    fld     st
                    add     esi,1

      @@20:         //==========================================================
                    // store bit_removed sample value and calculate
                    // and store sample correction value
                    //==========================================================
                    fist    dword ptr btrd_codec_block[ebx]
                    fmul    settings.scaling_factor_inv
                    fisubr  dword ptr [ebp+ebx]
                    fistp   dword ptr corr_codec_block[ebx]
                    //==========================================================

                    add     ecx,DOUBLE_SIZE
                    add     ebx,INTEGER_SIZE
                    cmp     ebx,edi
                    jl      @@09

                  cmp     esi,settings.maximum_clips_per_channel
                  jbe     @@24

                  fstp    st
                  fstp    st
                  fstp    st
                  fstp    st
                  sub     dword ptr remove_bits_data.bits_removed,1
                  jmp     @@01

      @@24:     mov     remove_bits_data.clips,esi
                fstp    st
                fstp    st
                fstp    st
                fstp    st

      @@29:     cmp     remove_bits_data.bits_removed,0
                jge     @@30

                mov     remove_bits_data.bits_removed,0

      @@30:     pop     edi
                pop     esi
                pop     ebp
                pop     ebx
  end;
  //============================================================================

  //============================================================================
  procedure remove_bits_proc_shaping_off; assembler;
  //============================================================================
  asm
                push    ebx
                push    ebp
                push    esi
                push    edi

      @@01:       mov     eax,dword ptr remove_bits_data.bits_removed
                  cmp     eax,0
                  jl      @@29

                  sub     esi,esi
                  fld     dword ptr PowersOf.Two[128*SINGLE_SIZE-eax*SINGLE_SIZE]
                  neg     eax
                  fld     dword ptr PowersOf.two[128*SINGLE_SIZE+eax*SINGLE_SIZE]
                  neg     eax
                  fild    dword ptr max_samples[eax*INTEGER_SIZE]
                  fild    this_min_sample

                  mov     ebx,Global.channel
                  mov     ebx,dword ptr channel_offset[ebx*POINTER_SIZE]
                  mov     ebp,dword ptr rotating_blocks_ptr[2*POINTER_SIZE]

                  mov     edi,this_codec_block_size
                  shl     edi,integer_shift
                  add     edi,ebx
                  mov     ecx,shaping_length*DOUBLE_SIZE

      @@09:         //==========================================================
                    // Start of bit-removal
                    //==========================================================
      @@10:         fild    dword ptr [ebp+ebx]                         // top of FPU stack = current sample value
                    fmul    settings.scaling_factor

                    //==========================================================
                    // Non-noise-shaped bit-removal.
                    //==========================================================
                    fmul    st,st(3)                                    // divide by 2^(bits-to-remove)
                    sub     esp,4
                    fistp   dword ptr [esp]                             // y[k] = round(wanted_temp); frndint replacement
                    fild    dword ptr [esp]                             // code - thanks Agner Fog!
                    add     esp,4
                    fmul    st,st(4)                                    // multiply by 2^(bits-to-remove)

      @@16:         //==========================================================
                    // Check that new value lies within acceptable limits.
                    //==========================================================
                    fcom    st(2)                 // max_sample
                    fnstsw  ax
                    and     ah,$41
                    jnz     @@18

                    fstp    st
                    fld     st(1)
                    add     esi,1
                    jmp     @@20

      @@18:         fcom    st(1)                 // min_sample
                    fnstsw  ax
                    xor     ah,$01
                    and     ah,$41
                    jnz     @@20

                    fstp    st
                    fld     st
                    add     esi,1

      @@20:         //==========================================================
                    // store bit_removed sample value and calculate
                    // and store sample correction value
                    //==========================================================
                    fist    dword ptr btrd_codec_block[ebx]
                    fmul    settings.scaling_factor_inv
                    fisubr  dword ptr [ebp+ebx]
                    fistp   dword ptr corr_codec_block[ebx]
                    //==========================================================

                    add     ecx,DOUBLE_SIZE
                    add     ebx,INTEGER_SIZE
                    cmp     ebx,edi
                    jl      @@09

                  cmp     esi,settings.maximum_clips_per_channel
                  jbe     @@24

                  fstp    st
                  fstp    st
                  fstp    st
                  fstp    st
                  sub     dword ptr remove_bits_data.bits_removed,1
                  jmp     @@01

      @@24:     fstp    st
                fstp    st
                fstp    st
                fstp    st
                mov     remove_bits_data.clips,esi

      @@29:     cmp     remove_bits_data.bits_removed,0
                jge     @@30

                mov     remove_bits_data.bits_removed,0

      @@30:     pop     edi
                pop     esi
                pop     ebp
                pop     ebx
  end;
  //============================================================================
  {$ENDIF}
  //============================================================================
  procedure fifo_save_state();
  //============================================================================
  var
    nrb_i, nrb_j : integer;
  begin
    for nrb_j:=0 to Global.channels-1 do
      for nrb_i:=0 to shaping_length-1 do
        saved_fifo_state[nrb_j,nrb_i]:=temp_fifo_state[nrb_j,nrb_i];
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure remove_bits();
  //============================================================================
  var
    nrb_i : integer;
  begin
    remove_bits_data.bits_removed:=remove_bits_data.bits_to_remove;

    for nrb_i:=0 to shaping_length-1 do
      fifo[-shaping_length+nrb_i]:=saved_fifo_state[Global.channel,nrb_i];

    if settings.shaping_is_on then
      remove_bits_proc_shaping_on
    else
      remove_bits_proc_shaping_off;

    for nrb_i:=0 to shaping_length-1 do
      temp_fifo_state[Global.channel,nrb_i]:=fifo[global.codec_block_size-shaping_length+nrb_i];
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure Remove_Bits_Init();
  //============================================================================
  var
    fbi_i, fbi_j : Integer;
  begin
    // Initialise noise shaping FIFO saved state information for each channel.
    for fbi_i:=0 to MAX_CHANNELS-1 do
      for fbi_j:=0 to SHAPING_LENGTH-1 do
        saved_fifo_state[fbi_i,fbi_j]:=0;

    this_max_sample:=PowersOf.twom1[bitspersample-1];
    this_min_sample:=-PowersOf.twoint[bitspersample-1];

    for fbi_i:=0 to BITS_TO_CALCULATE-1 do
      max_samples[fbi_i]:=(this_max_sample xor powersof.twom1[fbi_i]);

    for fbi_i:=0 to shaping_length-1 do
      shaping_factor_power[fbi_i]:=power(settings.shaping_factor,fbi_i+1);

    if samplerate>46050 then
      for fbi_i:=0 to shaping_length-1 do
      begin
        Shaping_Array.a[fbi_i]:=order_4_48000.a[fbi_i]*shaping_factor_power[fbi_i];
        Shaping_Array.b[fbi_i]:=order_4_48000.b[fbi_i]*shaping_factor_power[fbi_i];
      end
    else
      for fbi_i:=0 to shaping_length-1 do
      begin
        Shaping_Array.a[fbi_i]:=order_4_44100.a[fbi_i]*shaping_factor_power[fbi_i];
        Shaping_Array.b[fbi_i]:=order_4_44100.b[fbi_i]*shaping_factor_power[fbi_i];
      end;

    settings.static_maximum_bits_to_remove:=bitspersample-STATIC_MINIMUM_BITS_TO_KEEP;
  //============================================================================
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('nRemoveBits',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
