//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008, 2009 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nShiftBlocks;
//==============================================================================
interface
//==============================================================================
var
  Unit_Data_Packer1 : Array[0..15] of Integer;
//==============================================================================
Procedure Shift_Codec_Blocks();
//==============================================================================
implementation
//==============================================================================
uses
  nMaths,
  nParameter,
  nSystem,
  nWav;
var
  Unit_Data_Packer2 : Array [0..14] of Integer;

//==============================================================================
 {$IFNDEF USEASM}
  //----------------------------------------------------------------------------
  Procedure Shift_Codec_Blocks();
  //----------------------------------------------------------------------------
  var
   sc_i : Integer;
   sc_p : Pointer;
   channel : Integer;
   left_sample, right_sample : Double;

    //--------------------------------------------------------------------------
    Function Channel_RMS(this_channel : Integer) : Extended;
    //--------------------------------------------------------------------------
    var
      Channel_Temp : Double;
      sc_i : Integer;
    Begin
      channel_temp:=0;
      for sc_i:=0 to this_codec_block_size-1 do
        channel_temp:=channel_temp+sqr(1.0*rotating_blocks_ptr[2]^[this_channel,sc_i]);
      channel_RMS:=fastl2root(channel_temp*this_codec_block_size_recip);
    //--------------------------------------------------------------------------
    End;
    //--------------------------------------------------------------------------

  begin
    sc_p:=rotating_blocks_ptr[0];
    rotating_blocks_ptr[0]:=rotating_blocks_ptr[1];
    rotating_blocks_ptr[1]:=rotating_blocks_ptr[2];
    rotating_blocks_ptr[2]:=rotating_blocks_ptr[3];
    rotating_blocks_ptr[3]:=sc_p;

    prev_codec_block_size:=last_codec_block_size;
    last_codec_block_size:=this_codec_block_size;
    this_codec_block_size:=next_codec_block_size;
    next_codec_block_size:=0;

    if this_codec_block_size>0 then
    begin
      this_codec_block_size_recip:=OneOver[this_codec_block_size];
      inc(Global.blocks_processed);
      Global.blocks_processed_recip:=1/Global.blocks_processed;

      for channel:=0 to Global.channels-1 do
        this_codec_block_channel_log2rms[channel]:=Channel_RMS(channel);

      //========================================================================
      // optional examination of mid-side content.
      //========================================================================
      if midside and (Global.channels=2) then
      begin
        for sc_i:=0 to this_codec_block_size-1 do
        begin
          left_sample:=rotating_blocks_ptr[2]^[0,sc_i];
          right_sample:=rotating_blocks_ptr[2]^[1,sc_i];

          rotating_blocks_ptr[2]^[2,sc_i]:=round(single_half*(left_sample+right_sample));
          rotating_blocks_ptr[2]^[3,sc_i]:=round(single_half*(left_sample-right_sample));
        end;

        this_codec_block_channel_log2rms[2]:=Channel_RMS(2);
        this_codec_block_channel_log2rms[3]:=Channel_RMS(3);
      end;
    end;
  end;
  //----------------------------------------------------------------------------
  {$ELSE}
  //============================================================================
  Procedure Shift_Codec_Blocks();
  //============================================================================
  asm
                push    ebp
                push    ebx
                push    edi
                push    esi

                mov     eax,dword ptr rotating_blocks_ptr[3*POINTER_SIZE]
                xchg    eax,dword ptr rotating_blocks_ptr[2*POINTER_SIZE]
                xchg    eax,dword ptr rotating_blocks_ptr[1*POINTER_SIZE]
                xchg    eax,dword ptr rotating_blocks_ptr[0*POINTER_SIZE]
                mov     dword ptr rotating_blocks_ptr[3*POINTER_SIZE],eax

                sub     eax,eax
                xchg    eax,next_codec_block_size
                xchg    eax,this_codec_block_size
                xchg    eax,last_codec_block_size
                mov     prev_codec_block_size,eax

                mov     eax,this_codec_block_size
                cmp     eax,1
                jl      @@99

                fld     qword ptr OneOver[eax*DOUBLE_SIZE]
                fstp    this_codec_block_size_recip

                add     Global.blocks_processed,1
                fld1
                fidiv   Global.blocks_processed
                fstp    Global.blocks_processed_recip

                mov     ebx,Global.channels
                mov     ebp,dword ptr rotating_blocks_ptr[2*POINTER_SIZE]

      @@00:     sub     ebx,1
                jl      @@04

                  fldz
                  mov     ecx,dword ptr CHANNEL_OFFSET[ebx*POINTER_SIZE]
                  add     ecx,ebp
                  mov     edx,eax

      @@01:       sub     edx,1
                  jl      @@03

      @@02:       fild    dword ptr [ecx+edx*INTEGER_SIZE]
                  fmul    st,st
                  fadd
                  jmp     @@01

      @@03:     fmul    this_codec_block_size_recip
                call    fastlog2root
                fstp    qword ptr this_codec_block_channel_log2rms[ebx*DOUBLE_SIZE]
                jmp     @@00

      @@04:     call    midside
                test    eax,1
                jz      @@99

                cmp     Global.channels,2
                jne     @@99

                fldz
                fldz

                mov     ecx,ebp
                add     ecx,dword ptr CHANNEL_OFFSET[1*INTEGER_SIZE]
                mov     edi,ebp
                add     edi,dword ptr CHANNEL_OFFSET[2*INTEGER_SIZE]
                mov     esi,ebp
                add     esi,dword ptr CHANNEL_OFFSET[3*INTEGER_SIZE]

                mov     eax,this_codec_block_size

      @@05:     sub     eax,1
                jl      @@09

                fild    dword ptr [ebp+eax*INTEGER_SIZE]
                fiadd   dword ptr [ecx+eax*INTEGER_SIZE]
                fmul    SINGLE_HALF
                frndint
                fist    dword ptr [edi+eax*INTEGER_SIZE]
                fmul    st,st
                fadd

                fild    dword ptr [ebp+eax*INTEGER_SIZE]
                fisub   dword ptr [ecx+eax*INTEGER_SIZE]
                fmul    SINGLE_HALF
                frndint
                fist    dword ptr [esi+eax*INTEGER_SIZE]
                fmul    st,st
                fadd    st,st(2)
                fstp    st(2)

                jmp     @@05

      @@09:     fmul    this_codec_block_size_recip
                call    fastlog2root
                fstp    qword ptr this_codec_block_channel_log2rms[2*double_size]

                fmul    this_codec_block_size_recip
                call    fastlog2root
                fstp    qword ptr this_codec_block_channel_log2rms[3*double_size]

      @@99:     //fwait
                pop     esi
                pop     edi
                pop     ebx
                pop     ebp
  end;
  //============================================================================
  {$ENDIF}

 //==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer1[0]:=0;
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers('nShiftBlocks',@Unit_Data_Packer1,pointer(cardinal(@Unit_Data_Packer2)+sizeof(Unit_Data_Packer2)));
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
